//	Copyright (c) Des Herriott 1993, 1994
//	              Erik Kunze   1995, 1996
//	              Jens Kilian  1998
//
//	Permission to use, copy, modify, distribute, and sell this software and its
//	documentation for any purpose is hereby granted without fee, provided that
//	the above copyright notice appear in all copies and that both that
//	copyright notice and this permission notice appear in supporting
//	documentation, and that the name of the copyright holder not be used in
//	advertising or publicity pertaining to distribution of the software without
//	specific, written prior permission.  The copyright holder makes no
//	representations about the suitability of this software for any purpose.  It
//	is provided "as is" without express or implied warranty.
//
//	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
//	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
//	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
//	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
//	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
//	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//	PERFORMANCE OF THIS SOFTWARE.

//	Command-line configuration options.

// ARRGH. getopt.h assumes that C++ == ANSI C. It is not so.
#ifndef __STDC__
#  define __STDC__ 1
#  include <getopt.h>
#  undef __STDC__
#else
#pragma error "__STDC__ defined in C++ code - whodunit?"
#endif

#include <cstdio>
#include <cstring>

#include <byteorder.h>
#include <FindDirectory.h>
#include <Path.h>

#include "ZXConfig.h"
#include "screen.h"
#include "util.h"
#include "version.h"

// Static class members.
char *ZXConfig::smpProgName = 0;
ZXConfig *ZXConfig::smpConf = 0;

// Name of saved configuration file.
const char kSettingsFilename[] = "BeZX_settings";

// Define an enumeration and a long option table for all the options
// we understand.
#define OPTION_LIST(item)								\
	/* Standard options */								\
	item(help,        no_argument,       = 'h')	\
	item(version,     no_argument,       = 'v')	\
	item(libDir,      required_argument, = 'L')	\
																\
	/* Display options */								\
	item(location,    required_argument, = 128)	\
	item(mono,        no_argument,       )			\
	item(kbdImage,    required_argument, )			\
	item(kbdLocation, required_argument, )			\
	item(kbdSize,     required_argument, )			\
	item(kbdShown,    no_argument,       )			\
																\
	/* Screen options */									\
	item(scale,       required_argument, )			\
	item(flashing,    required_argument, )			\
	item(refresh,     required_argument, )			\
																\
	/* Emulation options */								\
	item(debug,       required_argument, )			\
	item(machine,     required_argument, )			\
	item(cycles,      required_argument, )			\
	item(ms,          required_argument, )			\
	item(slowdown,    required_argument, )			\
	item(keyFaster,   required_argument, )			\
	item(keySlower,   required_argument, )			\
																\
	/* Joystick options */								\
	item(joyUp,       required_argument, )			\
	item(joyDown,     required_argument, )			\
	item(joyLeft,     required_argument, )			\
	item(joyRight,    required_argument, )			\
	item(joyFire,     required_argument, )			\
	item(joystick,    required_argument, )			\
	item(joyDev,      required_argument, )			\
	item(tolerance,   required_argument, )			\
																\
	/* Sound options */									\
	item(sound,       required_argument, )			\
																\
	/* Spectrum 48k options */							\
	item(rom48,       required_argument, )			\
	item(issue,       required_argument, )			\
																\
	/* Spectrum 128k options */						\
	item(rom1280,     required_argument, )			\
	item(rom1281,     required_argument, )			\
																\
	/* Spectrum +3 options */							\
	item(rompl30,     required_argument, )			\
	item(rompl31,     required_argument, )			\
	item(rompl32,     required_argument, )			\
	item(rompl33,     required_argument, )			\
	item(fda,         required_argument, )			\
	item(fdb,         required_argument, )			\
																\
	/* Interface 1 options */							\
	item(if1,         required_argument, )			\
	item(if1Rom,      required_argument, )			\
	item(crlf,        required_argument, )			\
	item(strcr,       required_argument, )			\
	item(m1,          required_argument, )			\
	item(m2,          required_argument, )			\
	item(m3,          required_argument, )			\
	item(m4,          required_argument, )			\
	item(m5,          required_argument, )			\
	item(m6,          required_argument, )			\
	item(m7,          required_argument, )			\
	item(m8,          required_argument, )			\
																\
	/* Multiface 128 options */						\
	item(mf128,       required_argument, )			\
	item(mf128Rom,    required_argument, )

#define OPTION_ENUM(name, arg, init)				\
	kO ## name init,

enum {
	OPTION_LIST(OPTION_ENUM)
	kNumOptions
};

#define OPTION_ENTRY(name, arg, init)				\
	{ #name, arg, 0, kO ## name },

static const option kOptions[] = {
	OPTION_LIST(OPTION_ENTRY)
	{ 0, 0, 0, 0 }
};

// Names corresponding to BeOS key codes.
static const char *kpKeyNames[128] = {
	0,							// 0x00
	"escape",
	"f1",
	"f2",
	"f3",
	"f4",
	"f5",
	"f6",
	"f7",
	"f8",
	"f9",
	"f10",
	"f11",
	"f12",
	"printscreen",
	"scrolllock",
	"pause",					// 0x10
	"grave",
	"1",
	"2",
	"3",
	"4",
	"5",
	"6",
	"7",
	"8",
	"9",
	"0",
	"minus",
	"equal",
	"backspace",
	"insert",
	"home",					// 0x20
	"pageup",
	"numlock",
	"kp_divide",
	"kp_multiply",
	"kp_subtract",
	"tab",
	"q",
	"w",
	"e",
	"r",
	"t",
	"y",
	"u",
	"i",
	"o",
	"p",						// 0x30
	"bracketleft",
	"bracketright",
	"backslash",
	"delete",
	"end",
	"pagedown",
	"kp_7",
	"kp_8",
	"kp_9",
	"kp_add",
	"capslock",
	"a",
	"s",
	"d",
	"f",
	"g",						// 0x40
	"h",
	"j",
	"k",
	"l",
	"semicolon",
	"apostrophe",
	"return",
	"kp_4",
	"kp_5",
	"kp_6",
	"shiftleft",
	"z",
	"x",
	"c",
	"v",
	"b",						// 0x50
	"n",
	"m",
	"comma",
	"period",
	"slash",
	"shiftright",
	"up",
	"kp_1",
	"kp_2",
	"kp_3",
	"kp_enter",
	"controlleft",
	"altleft",
	"space",
	"altright",
	"controlright",		// 0x60
	"left",
	"down",
	"right",
	"kp_0",
	"kp_decimal",
	"optionleft",			// The following are not always present.
	"optionright",
	"menu",
	"euro",
	"kp_equal",
	"power",
	0,
	0,
	0,							// 0x70
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	"sysrq",
	"break"
};
const int kNumKeys = sizeof(kpKeyNames) / sizeof(kpKeyNames[0]);

// Default action keys.
static const char *kpDefaultActions[NUM_KEYMAPPINGS] = {
	"q",						// The time-honored game controls.
	"a",
	"o",
	"p",
	"space",
#ifdef SLOWDOWN
	"kp_add",
	"kp_subtract"
#endif
};

// Copy a string.
static inline char *
copyString(const char *pStr)
{
	return strcpy(new char[strlen(pStr)+1], pStr);
}

// Process single command line arguments.
static void
getString(const char *pArg, char *&pValue)
{
	delete [] pValue;
	pValue = copyString(pArg);
}

static void
getBool(const char *pArg, bool &value)
{
	value = !strcasecmp(pArg, "true")
		  || !strcasecmp(pArg, "on")
		  || !strcasecmp(pArg, "yes");
}

static void
getInteger(const char *pArg, int &value)
{
	int newValue;
	if (sscanf(pArg, "%d", &newValue) == 1) {
		value = newValue;
	}
}

static void
getKey(const char *pArg, uint8 &value)
{
	for (uint8 i = 0; i < kNumKeys; ++i) {
		if (kpKeyNames[i] && !strcasecmp(pArg, kpKeyNames[i])) {
			value = i;
			return;
		}
	}
}

static void
getPoint(const char *pArg, BPoint &value)
{
	float x, y;
	if (sscanf(pArg, "%f,%f", &x, &y) == 2) {
		value = BPoint(x, y);
	}
}

// Print out a copyright message.
static void
copyright(void)
{
	fprintf(stderr,
	"BeZX is derived from Xzx, which is copyright (c) 1993,1994 Des Herriott\n"
	"                                             (c) 1995,1996 Erik Kunze.\n"
	"BeOS specific code copyright (c) 1998 Jens Kilian.\n"
	"BeZX comes with no warranty; see the file COPYRIGHT for details.\n");
}

// Print out a usage message.
static void
usage(void)
{
	fprintf(stderr, "Usage: %s [options] [snapshot file]\n", ProgName);
	fprintf(stderr, "Available options:\n"

	/* Standard options */
	"  -h, --help                 display this message\n"
	"  -v, --version              display version number and options\n"

	"  -L <path>, --libDir <path> specify library search path\n"

	/* Display options */
	"  --location <X,Y>           specify window position\n"
	"  --mono                     use only black and white\n"
	"  --kbdImage <name>          image file for keyboard window\n"
	"  --kbdLocation <X,Y>        specify keyboard window position\n"
	"  --kbdSize <X,Y>            specify keyboard window size\n"
	"  --kbdShown                 show keyboard window\n"

	/* Screen options */
	"  --scale <n>                Spectrum to BeOS screen scaling factor\n"
	"  --flashing <bool>          enable flashing attributes\n"
	"  --refresh <n>              refresh screen every <n> frames\n"

	/* Emulation options */
#ifdef DEBUG
	"  --debug <n>                specify the debug level\n"
#endif
	"  --machine [48|128|3]       emulate the specified Spectrum type\n"
	"  --cycles <n>               specify number of T-states per frame\n"
	"  --ms <n>                   specify number of milliseconds per frame\n"
#ifdef SLOWDOWN
	"  --slowdown <int>           specify slowdown factor\n"
	"  --keyFaster <key>          define key to speed up the emulator\n"
	"  --keySlower <key>          define key to slow down the emulator\n"
#endif

	/* Joystick options */
	"  --joyUp <key>              define key for joystick up\n"
	"  --joyDown <key>            define key for joystick down\n"
	"  --joyLeft <key>            define key for joystick left\n"
	"  --joyRight <key>           define key for joystick right\n"
	"  --joyFire <key>            define key for joystick fire\n"
	"  --joystick <bool>          enable/disable Kempston joystick\n"
	"  --joyDev <name>            specify joystick device\n"
	"  --tolerance <int>          specify joystick displacement tolerance\n"

	/* Sound options */
	"  --sound <bool>             enable/disable sound\n"

	/* Spectrum 48k options */
	"  --rom48 <fname>            specify the file to use as 48K ROM image\n"
	"  --issue [2|3]              emulate specified issue on 48K Spectrums\n"

	/* Spectrum 128k options */
	"  --rom128{0,1} <fname>      use the specified file as 128K ROM image\n"

	/* Spectrum +3 options */
	"  --rompl3{0-3} <fname>      specify the file to use as +3 ROM image\n"
	"  --fd[ab] <fname>           specify file to use as virtual disk drive\n"

	/* Interface 1 options */
	"  --if1 <bool>               enable Interface 1\n"
	"  --if1Rom <fname>           specify the file to use as IF1 ROM image\n"
	"  --crlf <bool>              enable CR/LF translation on RS232 input\n"
	"  --strcr <bool>             strip CR on RS232 output\n"
	"  --m<n> <fname>             put Microdrive cartridge file in drive <n>\n"

	/* Multiface 128 options */
	"  --mf128 <bool>             enable Multiface 128\n"
	"  --mf128Rom <fname>         specify the file to use as MF128 ROM image\n"

	"\nMicrodrive emulation reads and writes .MDR files.\n"
	"BeZX reads and writes .SNA & .Z80 format snapshots.\n");
}

// Print out the version information.
static void
version(void)
{
	copyright();
	
	fprintf(stderr, "\nThis is BeZX, based on Xzx %d.%d.%d",
			  ZX_Major, ZX_Minor, ZX_Patch);
	fprintf(stderr,	" with the following compile time options:\n"

	" *  BeZX is built for a"
#if B_HOST_IS_LENDIAN
	" little-endian architecture.\n"
#else
	" big-endian architecture.\n"
#endif
#ifdef DEBUG
    " *  Debugging is enabled.\n"
#endif
#ifdef SLOWDOWN
    " *  Speed control is enabled.\n"
#endif
	);
}

// Constructor - initializes and parses command-line options.
ZXConfig::ZXConfig(int32 &argc, char **ppArgv, const char *pAppDir)
:	libDir(0),
	location(100, 100),
	mono(false),
	kbdImage(copyString("ZXKeyboard.tiff")),
	kbdLocation(300, 300),
	kbdSize(0, 0),
	kbdShown(false),
	scale(1),
	flashing(true),
	refreshRate(2),
#ifdef DEBUG
	debug(0),
#endif
	machine(48),
	cycles(69888),
	ms(20),
#ifdef SLOWDOWN
	slowdown(0),
#endif
	useJoy(true),
	joyDevice(copyString("joystick1")),
	joyTolerance(20),
	useSound(false),
	rom48(copyString("spectrum.rom")),
	issue(3),
	rom128_0(copyString("128-0.rom")),
	rom128_1(copyString("128-1.rom")),
	rompl3_0(copyString("pl3-0.rom")),
	rompl3_1(copyString("pl3-1.rom")),
	rompl3_2(copyString("pl3-2.rom")),
	rompl3_3(copyString("pl3-3.rom")),
	disk_a(copyString("disk-a.dsk")),
	disk_b(copyString("disk-b.dsk")),
	if1_active(true),
	if1_rom(copyString("if1.rom")),
	translate_nl(false),
	strip_nl(false),
	mf128_active(false),
	mf128_rom(copyString("mf128.rom"))
{
	// Initialize program name.
	if ((smpProgName = strrchr(ppArgv[0], '/')) != NULL) {
		++smpProgName;
	} else {
		smpProgName = ppArgv[0];
	}
	
	// Initialize search path.
	if (pAppDir) {
		const size_t appDirLength = strlen(pAppDir);
		const int libDirLength =
			2 * appDirLength + 14;		// "<appDir>:<appDir>/lib:.:./lib\0"
		libDir = new char[libDirLength];
		sprintf(libDir, "%s:%s/lib:.:./lib", pAppDir, pAppDir);

	} else {
		libDir = copyString(".:./lib");
	}
	
	// Initialize action keys.
	for (int i = 0; i < NUM_KEYMAPPINGS; ++i) {
		getKey(kpDefaultActions[i], actionKeys[i]);
	}

	// Initialize cartridge file names.	
	for (int i = 0; i < 8; i++) {
		char cartridge[5];
		sprintf(cartridge, "MDR%d", i + 1);
		cartfiles[i] = copyString(cartridge);
	}

	// Try to load saved settings.
	LoadSettings();
	
	bool errorSeen = false;
	int c;
	while ((c = getopt_long(argc, ppArgv, "hvL:", kOptions, 0)) != EOF) {
		if (!ProcessOption(c, optarg)) {
			errorSeen = true;
		}
	}
	
	// Check consistency of settings.
	if (machine != 48 && machine != 128 && machine != 3) {
		XzxMesg(XZX_WARN, "invalid emulation mode %d, using 48", machine);
		machine = 48;
		errorSeen = true;
	}
	
	if (ms <= 0) {
		XzxMesg(XZX_WARN, "timer value %d out of range, setting to default 20", ms);
		ms = 20;
		errorSeen = true;
	}

	if (scale < 1 || scale > 3) {
		XzxMesg(XZX_WARN, "scale factor out of range, setting to default 1");
		scale = 1;
		errorSeen = true;
	}

	if (issue != 2 && issue != 3) {
		XzxMesg(XZX_WARN, "invalid issue type %d, using 3", issue);
		issue = 3;
		errorSeen = true;
	}

	if (machine == 3) {
		if (if1_active) {
			XzxMesg(XZX_INFO, "Interface 1 not supported by the +3, disabling");
			if1_active = false;
			errorSeen = true;
		}
		if (mf128_active) {
			XzxMesg(XZX_INFO, "Multiface 128 not supported by the +3, disabling");
			mf128_active = false;
			errorSeen = true;
		}
	}
	
	if (errorSeen) {
		fprintf(stderr, "Type: %s -help to see a list of options.\n", ppArgv[0]);
	}

	// Remove all the parsed options from the command line.
	argc -= optind - 1;
	for (int i = 1; i < argc; ++i) {
		ppArgv[i] = ppArgv[i + optind - 1];
	}
}

// Destructor.
ZXConfig::~ZXConfig(void)
{
	// Try to save current settings.
	SaveSettings();

	// Free memory owned by us.
	delete [] libDir;
	delete [] kbdImage;
	delete [] joyDevice;
	delete [] rom48;
	delete [] rom128_0;
	delete [] rom128_1;
	delete [] rompl3_0;
	delete [] rompl3_1;
	delete [] rompl3_2;
	delete [] rompl3_3;
	delete [] disk_a;
	delete [] disk_b;
	delete [] if1_rom;
	for (int i = 0; i < 8; ++i) {
		delete [] cartfiles[i];
	}
	delete [] mf128_rom;
}

// Process a single configuration option.
bool
ZXConfig::ProcessOption(int option, const char *pArg)
{
	switch (option) {

	case kOhelp:
		usage();
		break;

	case kOversion:
		version();
		break;

	case kOlibDir:
		getString(pArg, libDir);
		break;

	case kOlocation:
		getPoint(pArg, location);
		break;

	case kOmono:
		mono = true;
		break;

	case kOkbdImage:
		getString(pArg, kbdImage);
		break;

	case kOkbdLocation:
		getPoint(pArg, kbdLocation);
		break;

	case kOkbdSize:
		getPoint(pArg, kbdSize);
		break;

	case kOkbdShown:
		kbdShown = true;
		break;

	case kOscale:
		getInteger(pArg, scale);
		break;

	case kOflashing:
		getBool(pArg, flashing);
		break;

	case kOrefresh:
		getInteger(pArg, refreshRate);
		break;
#ifdef DEBUG
	case kOdebug:
		getInteger(pArg, debug);
		break;
#endif
	case kOmachine:
		getInteger(pArg, machine);
		break;

	case kOcycles:
		getInteger(pArg, cycles);
		break;

	case kOms:
		getInteger(pArg, ms);
		break;
#ifdef SLOWDOWN
	case kOslowdown:
		getInteger(pArg, slowdown);
		break;

	case kOkeyFaster:
		getKey(pArg, actionKeys[KM_KEY_FASTER]);
		break;

	case kOkeySlower:
		getKey(pArg, actionKeys[KM_KEY_SLOWER]);
		break;
#endif
	case kOjoyUp:
		getKey(pArg, actionKeys[KM_JOY_UP]);
		break;

	case kOjoyDown:
		getKey(pArg, actionKeys[KM_JOY_DOWN]);
		break;

	case kOjoyLeft:
		getKey(pArg, actionKeys[KM_JOY_LEFT]);
		break;

	case kOjoyRight:
		getKey(pArg, actionKeys[KM_JOY_RIGHT]);
		break;

	case kOjoyFire:
		getKey(pArg, actionKeys[KM_JOY_FIRE]);
		break;

	case kOjoystick:
		getBool(pArg, useJoy);
		break;

	case kOjoyDev:
		getString(pArg, joyDevice);
		break;

	case kOtolerance:
		getInteger(pArg, joyTolerance);
		break;

	case kOsound:
		getBool(pArg, useSound);
		break;

	case kOrom48:
		getString(pArg, rom48);
		break;

	case kOissue:
		getInteger(pArg, issue);
		break;

	case kOrom1280:
		getString(pArg, rom128_0);
		break;

	case kOrom1281:
		getString(pArg, rom128_1);
		break;

	case kOrompl30:
		getString(pArg, rompl3_0);
		break;

	case kOrompl31:
		getString(pArg, rompl3_1);
		break;

	case kOrompl32:
		getString(pArg, rompl3_2);
		break;

	case kOrompl33:
		getString(pArg, rompl3_3);
		break;

	case kOfda:
		getString(pArg, disk_a);
		break;

	case kOfdb:
		getString(pArg, disk_b);
		break;

	case kOif1:
		getBool(pArg, if1_active);
		break;

	case kOif1Rom:
		getString(pArg, if1_rom);
		break;

	case kOcrlf:
		getBool(pArg, translate_nl);
		break;

	case kOstrcr:
		getBool(pArg, strip_nl);
		break;

	case kOm1:
	case kOm2:
	case kOm3:
	case kOm4:
	case kOm5:
	case kOm6:
	case kOm7:
	case kOm8:
		getString(pArg, cartfiles[option - kOm1]);
		break;

	case kOmf128:
		getBool(pArg, mf128_active);
		break;

	case kOmf128Rom:
		getString(pArg, mf128_rom);
		break;
	
	default:
		// Error.
		return false;
	}

	// OK, no error.
	return true;
}

// Try to load resp. save configuration.
void
ZXConfig::LoadSettings(void)
{
	BPath settingsPath;
	FILE *pSettingsFile;
	if ((find_directory(B_USER_SETTINGS_DIRECTORY, &settingsPath) < B_OK
		  || settingsPath.Append(kSettingsFilename) < B_OK
		  || (pSettingsFile = fopen(settingsPath.Path(), "r")) == NULL)
		 && (find_directory(B_COMMON_SETTINGS_DIRECTORY, &settingsPath) < B_OK
			  || settingsPath.Append(kSettingsFilename) < B_OK
			  || (pSettingsFile = fopen(settingsPath.Path(), "r")) == NULL)) {
		// No readable settings file found.
		return;
	}
	
	// Read lines from the file until EOF is encountered.
	char linebuf[1024];
	char *pLine;
	while ((pLine = fgets(linebuf, sizeof(linebuf), pSettingsFile)) != NULL) {
		// Lines which are empty or start with '#' are ignored.
		// Format of other lines is simply name[=value].
		if (*pLine != '\n' && *pLine != '#') {
			char *pArg = strchr(pLine, '=');
			if (pArg) {
				*pArg++ = '\0';
			}
			char *pEnd = strchr(pArg ? pArg : pLine, '\n');
			if (pEnd) {
				*pEnd = '\0';
			}
			
			// Names of entries are the same as for the corresponding
			// command-line options.
			const option *pOption;
			for (pOption = kOptions; pOption->name; ++pOption) {
				if (!strcmp(pLine, pOption->name)) {
					break;
				}
			}
			if (!(pOption->name && ProcessOption(pOption->val, pArg))) {
				fprintf(stderr, "%s: unrecognized configuration option %s%s%s\n",
						  ProgName, pLine, pArg ? "=" : "", pArg ? pArg : "");
			}
		}
	}
	
	// OK, finished.
	fclose(pSettingsFile);
}

static void
stringOption(FILE *pSettingsFile, const char *pName, const char *pValue)
{
	if (pValue) {
		fprintf(pSettingsFile, "%s=%s\n", pName, pValue);
	}
}

static void
boolOption(FILE *pSettingsFile, const char *pName, bool value)
{
	fprintf(pSettingsFile, "%s=%s\n", pName, value ? "true" : "false");
}

static void
integerOption(FILE *pSettingsFile, const char *pName, int value)
{
	fprintf(pSettingsFile, "%s=%d\n", pName, value);
}

static void
keyOption(FILE *pSettingsFile, const char *pName, uint8 key)
{
	if (kpKeyNames[key]) {
		fprintf(pSettingsFile, "%s=%s\n", pName, kpKeyNames[key]);
	}
}

static void
pointOption(FILE *pSettingsFile, const char *pName, BPoint point)
{
	fprintf(pSettingsFile, "%s=%f,%f\n", pName, point.x, point.y);
}

void
ZXConfig::SaveSettings(void)
{
	BPath settingsPath;
	FILE *pSettingsFile;
	if ((find_directory(B_USER_SETTINGS_DIRECTORY, &settingsPath) < B_OK
		  || settingsPath.Append(kSettingsFilename) < B_OK
		  || (pSettingsFile = fopen(settingsPath.Path(), "w")) == NULL)) {
		// Could not open file for writing.
		return;
	}
	
	// Write out current settings.
	fprintf(pSettingsFile, "# Configuration options written by %s.\n\n",
			  ProgName);
	
	stringOption(pSettingsFile, "libDir", libDir);
	pointOption(pSettingsFile, "location", location);
	if (mono) {
		fprintf(pSettingsFile, "mono\n");
	}
	stringOption(pSettingsFile, "kbdImage", kbdImage);
	pointOption(pSettingsFile, "kbdLocation", kbdLocation);
	pointOption(pSettingsFile, "kbdSize", kbdSize);
	if (kbdShown) {
		fprintf(pSettingsFile, "kbdShown\n");
	}
	integerOption(pSettingsFile, "scale", scale);
	boolOption(pSettingsFile, "flashing", flashing);
	integerOption(pSettingsFile, "refresh", refreshRate);
#ifdef DEBUG
	integerOption(pSettingsFile, "debug", debug);
#endif
	integerOption(pSettingsFile, "machine", machine);
	integerOption(pSettingsFile, "cycles", cycles);
	integerOption(pSettingsFile, "ms", ms);
#ifdef SLOWDOWN
	integerOption(pSettingsFile, "slowdown", slowdown);
	keyOption(pSettingsFile, "keyFaster", actionKeys[KM_KEY_FASTER]);
	keyOption(pSettingsFile, "keySlower", actionKeys[KM_KEY_SLOWER]);
#endif
	keyOption(pSettingsFile, "joyUp", actionKeys[KM_JOY_UP]);
	keyOption(pSettingsFile, "joyDown", actionKeys[KM_JOY_DOWN]);
	keyOption(pSettingsFile, "joyLeft", actionKeys[KM_JOY_LEFT]);
	keyOption(pSettingsFile, "joyRight", actionKeys[KM_JOY_RIGHT]);
	keyOption(pSettingsFile, "joyFire", actionKeys[KM_JOY_FIRE]);
	boolOption(pSettingsFile, "joystick", useJoy);
	stringOption(pSettingsFile, "joyDev", joyDevice);
	integerOption(pSettingsFile, "tolerance", joyTolerance);
	boolOption(pSettingsFile, "sound", useSound);
	stringOption(pSettingsFile, "rom48", rom48);
	integerOption(pSettingsFile, "issue", issue);
	stringOption(pSettingsFile, "rom1280", rom128_0);
	stringOption(pSettingsFile, "rom1281", rom128_1);
	stringOption(pSettingsFile, "rompl30", rompl3_0);
	stringOption(pSettingsFile, "rompl31", rompl3_1);
	stringOption(pSettingsFile, "rompl32", rompl3_2);
	stringOption(pSettingsFile, "rompl33", rompl3_3);
	stringOption(pSettingsFile, "fda", disk_a);
	stringOption(pSettingsFile, "fdb", disk_b);
	boolOption(pSettingsFile, "if1", if1_active);
	stringOption(pSettingsFile, "if1Rom", if1_rom);
	boolOption(pSettingsFile, "crlf", translate_nl);
	boolOption(pSettingsFile, "strcr", strip_nl);
	stringOption(pSettingsFile, "m1", cartfiles[0]);
	stringOption(pSettingsFile, "m2", cartfiles[1]);
	stringOption(pSettingsFile, "m3", cartfiles[2]);
	stringOption(pSettingsFile, "m4", cartfiles[3]);
	stringOption(pSettingsFile, "m5", cartfiles[4]);
	stringOption(pSettingsFile, "m6", cartfiles[5]);
	stringOption(pSettingsFile, "m7", cartfiles[6]);
	stringOption(pSettingsFile, "m8", cartfiles[7]);
	boolOption(pSettingsFile, "mf128", mf128_active);
	stringOption(pSettingsFile, "mf128Rom", mf128_rom);

	// OK, finished.
	fclose(pSettingsFile);
}
