// This code is from a tutorial by William Adams, and was only
// slightly rewritten.

#include <math.h>
#include <stdio.h>

#include <Joystick.h>
#include <SupportDefs.h>

#include "gamestick.h"

// Method: Constructor
//
// The constructor takes a parameter to specify
// which game port the user wants to connect to.
// Valid values are:
//		joystick1	- The lower joystick port; default.
//		joystick3	- The upper joystick port.
//		joystick2	- Second of lower port with splitter.
//		joystick4	- Second of upper port with splitter.

BTSGameStick::BTSGameStick(const char *pDeviceName)
:	mJoystick(),
	mIsValid(mJoystick.Open(pDeviceName) >= B_OK),
	mZeroX(0.0),
	mZeroY(0.0),
	mMinX(10000.0),
	mMinY(10000.0),
	mMaxX(-10000.0),
	mMaxY(-10000.0),
	mMinXRange(50.0),
	mMinYRange(50.0),
	mMaxXRange(50.0),
	mMaxYRange(50.0)
{
	float x, y;
	FindZero(x, y);
}

// Method: FindZero
//
// This method tries to find a good zero value for the
// joystick.  The zero values are the x and y values 
// that are returned when the joystick is at rest.  Thus
// this routine should be called whenever the stick is sure
// to be at rest.
//
// A number of consecutive samples are taken from the 
// port and a weighted average is maintained.  At the
// end of the cycle, the zero values should be set.
//
// After the zero values are established, the gamestick
// is trained to get the current range values.
void
BTSGameStick::FindZero(float &x, float &y)
{
	double totalX = 0.0;
	double totalY = 0.0;
	
	for (int counter = 1; counter <= 10; counter++) {
		mJoystick.Update();
		totalX = ((counter-1)*totalX + mJoystick.horizontal)/counter;
		totalY = ((counter-1)*totalY + mJoystick.vertical)/counter;
		//printf("totalX: %g  totalY: %g\n", totalX, totalY);
	}
	
	mZeroX = totalX;
	mZeroY = totalY;
	
	mLastX = mZeroX;
	mLastY = mZeroY;
	
	Train();

	x = mZeroX;
	y = mZeroY;
}

// Function: clampLow
//
// Make sure that the x value is not less than the
// specified low mark.  Since the joystick is self
// training, this helps to ensure that there aren't
// wild fluctuations in the range of values when the
// stick is first moved.
static inline void
clampLow(float &x, const float lowmark)
{
	if (x < lowmark) {
		x = lowmark;
	}
}

//	Method: Train
//
//	Training the joystick involves using the
//	the latest information to generate new x and
//	y min and max values.  This auto training allows
//	the joystick to learn its limits over time 
//	without having to tell the user to move the 
//	stick around to calibrate it in the beginning
//	of a game.  
//
//	Training does not find the zero value.  Here we
//	assume the zero position has already been found
//	by the constructor code, which assumes the stick 
//	is in a neutral position when it starts.
void
BTSGameStick::Train(void)
{
	if (mLastX < mMinX) {
		mMinX = mLastX;
	}
	if (mLastX > mMaxX) {
		mMaxX = mLastX;
	}

	if (mLastY < mMinY) {
		mMinY = mLastY;
	}
	if (mLastY > mMaxY) {
		mMaxY = mLastY;
	}
		
	mMinXRange = mZeroX - mMinX;
	mMinYRange = mZeroY - mMinY;

	mMaxXRange = mMaxX - mZeroX;
	mMaxYRange = mMaxY - mZeroY;
	
	// Create a good dead buffer zone.
	clampLow(mMinXRange, 50.0);
	clampLow(mMinYRange, 50.0);
	clampLow(mMaxXRange, 50.0);
	clampLow(mMaxYRange, 50.0);
}

// Method: Update
void
BTSGameStick::Update(void)
{
	float x, y;
	GetPosition(x,y);
}

// Method: GetPosition
//
// Returns x and y values from -1.0 to 1.0.
//
//	        1.0 y
//          ^
//          |
//	-1.0 <- 0.0 -> 1.0 x
//          |
//          V
//        -1.0
void
BTSGameStick::GetPosition(float &x, float &y)
{
	mJoystick.Update();
	
	mLastX = mJoystick.horizontal;
	mLastY = mJoystick.vertical;
	Train();
	
	float xOffset = mLastX - mZeroX;
	float yOffset = mLastY - mZeroY;
	
	if (xOffset < 0.0) {
		x = xOffset / mMinXRange;
	} else {
		x = xOffset / mMaxXRange;
	}
	x = -x;
	
	if (yOffset > 0.0) {
		y = yOffset / mMaxYRange;
	} else {
		y = yOffset / mMinYRange;
	}
	y = -y;
}

//	Method: GetAbsolutePosition
//
// This method returns the values as reported by the
// joystick port directly.  If you are going to use
// this method, then you may consider just using the
// BJoystick object directly.
void
BTSGameStick::GetAbsolutePosition(short &x, short &y)
{
	mJoystick.Update();
	
	x = mJoystick.horizontal;
	y = mJoystick.vertical;

	// Record this position and train again.
	mLastX = x;
	mLastY = y;
	Train();
}

//	Method: GetAbsolutePosition
//
// This method returns the calculated zero position.
void
BTSGameStick::GetZeroPosition(short &x, short &y)
{
	x = mZeroX;
	y = mZeroY;
}

//	Method: IsButtonPressed
//
//	The method tells you whether a particular button
//	is pressed or not.  The 'buttons' parameter
//	is a bitfield constructed by ORing together the
//	enums for kButton1, and kButton2.
//
//	Returns:
//		This returns either true or false for the
//		buttons that were passed in the 'buttons'
//		field.
//
//	Note: When you get raw joystick data, these 
//		values are the inverse of the logical button
//		state.  That is, when a button is pressed, 
//		the raw data will show a 0, when a button is
//		not pressed, it will show a 1.  This routine
//		returns the logical state, which is the opposite
//		of these raw numbers.
bool	
BTSGameStick::IsButtonPressed(const int buttons)
{
	mJoystick.Update();
	
	if (buttons & kButton1) {
		if (!mJoystick.button1)	{
			return true;
		}
	}
	
	if (buttons & kButton2) {
		if (!mJoystick.button2)	{
			return true;
		}
	}
	
	return false;
}
