/********************************************************************************
 * Copyright (c) Erik Kunze   1996
 *               Ville Hallik 1996
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *         Ville Hallik (basic AY-3-8912 emulation)
 *
 * changed by EKU
 *******************************************************************************/

static char ay8912_c[] = "$Id: ay8912.c,v 3.0 1996/10/03 10:37:58 erik Exp $";

/********************************************************************************
 * ay8912.c - AY-3-8912 sound emulation.
 * 
 * Currently we have:
 *     o /dev/dsp      (Linux w/USS)
 *     o /dev/audio    (Sun Sparc, Linux, NEC 4800/330)
 *******************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>

#if defined(__sun__)
#include <sun/audioio.h>
#endif

#if defined(__linux__) || defined(XXX)
#include <sys/ioctl.h>
#ifdef __linux__
#include <linux/soundcard.h>
#else
#include <machine/soundcard.h>
#endif
#endif

#include "z80.h"
#include "debug.h"
#include "util.h"
#include "ay8912.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

#define	SND_PAUSE		SIGUSR1
#define SND_RESUME		SIGUSR2
#define SND_QUIT		SIGHUP

/********************************************************************************
 * Forwards
 *******************************************************************************/

static void ayQuit(void);
static int ayOpenDevice(void);
static void aySignal(int);
static void ayPlay(void);

/********************************************************************************
 * Globals
 *******************************************************************************/

/********************************************************************************
 * Locals
 *******************************************************************************/

static int ayOn = 0;
static int ayActive = 0;
static int ayFd = -1;
static char *ayDevice;
static pid_t ayChild = 0;
static int ayParent = 1;
static int ayPaused = 0;
static int ayPipe[2] = { -1, -1 };
static int aySampleRate = SOUND_RATE;

static unsigned char envelopeForms[16][32] = {
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	  15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15 },
	{ 15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0,
	  15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	  15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	  15, 14, 13, 12, 11, 10,  9,  8,  7,  6,  5,  4,  3,  2,  1,  0 },
	{  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
	   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 }
};

/********************************************************************************
 * AyInit() - Initialize the AY emulation.
 *
 * Input:  audioDev - the audio device to use
 *         audioOn - turn AY on
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

int
AyInit(char *audioDev, int audioOn)
{
	OnQuit(ayQuit);

	ayDevice = audioDev;

	if (ayOpenDevice() == -1)
	{
		return -1;
	}
	
	/* Create a pipe for communication between parent and child process. */
	if (pipe(ayPipe) == -1)
	{
		XzxMesg(XZX_WARN, "couldn't create communication pipe");
		close(ayFd);
		ayFd = -1;
		return -1;
	}

	/* Install signal handler before (!) forking a child. */
	signal(SIGCHLD,    aySignal);
	signal(SIGPIPE,    aySignal);
	signal(SND_PAUSE,  aySignal);
	signal(SND_RESUME, aySignal);
	signal(SND_QUIT,   aySignal);
	
	/* Now fork a child, that makes the work. */
	switch (ayChild = fork())
	{
		case -1: /* failed */
			XzxMesg(XZX_WARN, "couldn't fork AY-3-8292 daemon");
			close(ayFd);
			ayFd = -1;
			return -1;
		case 0:  /* child */
			ayParent = 0;
			fcntl(ayPipe[0], F_SETFL, O_NONBLOCK);
			close(ayPipe[1]);
			ayPlay();
			break;
		default: /* parent */
			close(ayPipe[0]);
			fcntl(ayPipe[1], F_SETFL, O_NONBLOCK);
			close(ayFd);
			break;
	}

	ayActive = 1;
	ayOn = 1;
		
	if (!audioOn)
	{
		AyOnOff();
	}
	
	return 0;
}

/********************************************************************************
 * ayQuit() - Quit the AY emulation.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
ayQuit(void)
{
	int status;
	
	if (ayParent)
	{
		/* Remove the previous installed signal handler. This is mainly done
		   to avoid the message 'AY-3-8912 daemon died'. */
		signal(SIGCHLD, SIG_IGN);
		
		kill(ayChild, SND_QUIT);
		wait(&status);
	}
	else
	{
		if (ayFd != -1)
		{
			close(ayFd);
		}
	}

	if (ayPipe[0] != -1)
	{
		close(ayPipe[0]);
	}
	if (ayPipe[1] != -1)
	{
		close(ayPipe[1]);
	}
}

/********************************************************************************
 * ayOpenDevice() - Open the audio device for the AY emulation.
 *
 * Input:  -
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

static int
ayOpenDevice(void)
{
#if defined(__linux__) || defined(XXX)

	int i;
	
	if ((ayFd = open(ayDevice, O_WRONLY | O_NONBLOCK)) == -1)
	{
		XzxMesg(XZX_PWARN,
				"couldn't open audio device <%s> for writing", ayDevice);
		return -1;
	}

	/* Set sample resolution. */
	i = AFMT_U8;
	if (ioctl(ayFd, SNDCTL_DSP_SETFMT, &i) == -1) {
		XzxMesg(XZX_PWARN, "couldn't set sample resolution to 8 bit");
		close(ayFd);
		return -1;
	}

	/* Set sample rate. */
	if (ioctl(ayFd, SNDCTL_DSP_SPEED, &aySampleRate) == -1)
	{
		XzxMesg(XZX_WARN, "couldn't set sample rate to %d Hz", aySampleRate);
		close(ayFd);
		return -1;
	}

	/* Set number of buffers to use. */
	i = SOUND_BUFSIZE_BITS | (SOUND_NUM_OF_BUFS << 16);
	if (ioctl(ayFd, SNDCTL_DSP_SETFRAGMENT, &i) == -1)
	{
		XzxMesg(XZX_WARN, "couldn't set fragment size");
		close(ayFd);
		return -1;
	}

#else

	struct audio_info audioInf;
	
	if ((ayFd = open(ayDevice, O_WRONLY | O_NONBLOCK)) == -1)
	{
		XzxMesg(XZX_PWARN,
				"couldn't open audio device <%s> for writing", ayDevice);
		return -1;
	}

	/* Sun's specific initialization */
	AUDIO_INIT_INFO(&audioInf);
	audioInf.play.sample_rate = aySampleRate;
	audioInf.play.channels = 1;
	audioInf.play.precision = 8;
	audioInf.play.encoding = AUDIO_ENCODING_LINEAR;

	if (ioctl(ayFd, AUDIO_SETINFO, &audioInf))
	{
		XzxMesg(XZX_WARN, "couldn't set fragment size");
		close(ayFd);
		return -1;
	}

#endif /* __linux__ || XXX */

	return 0;
}
	
/********************************************************************************
 * AyOnOff - Turns AY sound on/off.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
AyOnOff(void)
{
	if (ayActive)
	{
		ayOn = !ayOn;

		if (ayChild > 0)  /* Any AY-3-8912 still daemon running? */
		{
			kill(ayChild, ayOn ? SND_RESUME : SND_PAUSE);
		}
	}
}

/********************************************************************************
 * aySignal() - Signal handler for parent/child process.
 *
 * Input:  signo - signal #
 *
 * Return: -
 *******************************************************************************/

static void
aySignal(int signo)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_AUDIO)
	{
		XzxMesg(XZX_DEBUG, "AY-3-8912: %s got signal %d",
				ayChild ? "parent" : "child", signo);
	}
#endif
	
	switch(signo)
	{
		/* Parent */
			
		case SIGPIPE:    /* Broken pipe */
			ayActive = 0;
			close(ayPipe[1]);
			break;
		case SIGCHLD:    /* Daemon died */
			ayActive = 0;
			ayChild = 0;
			XzxMesg(XZX_WARN, "AY-3-8912 daemon died");
			break;

		/* Child */
		
		case SND_PAUSE:  /* Pause daemon */
			signal(SND_PAUSE, aySignal);
			ayPaused = 1;
			close(ayFd);
			ayFd = -1;
			break;
		case SND_RESUME: /* Resume daemon */
			signal(SND_RESUME, aySignal);
			if (ayOpenDevice() != -1)
			{
				ayPaused = 0;
			}
			break;
		case SND_QUIT:  /* Shutdown daemon */
			close(ayPipe[0]);
			close(ayFd);
			exit(0);    /* Should this be a call to Quit()??? */
	}
}

/********************************************************************************
 * AyOutByte() - Output a byte to the AY-3-8219 sound chip.
 *
 * Input:  reg - register #
 *         val - register value
 *
 * Return: -
 *******************************************************************************/

void
AyOutByte(uns8 reg, uns8 val)
{
	uns8 buf[2];

	if (ayActive && ayOn)
	{
		buf[0] = reg;
		buf[1] = val;
		write(ayPipe[1], buf, sizeof(buf));
	}
}

/********************************************************************************
 * The AY-3-8910/8912 sound synthesizer IC.
 * ========================================
 *
 * Overview
 * --------
 * The AY-3-8910/8912 is a sound generator IC containing three tone generators
 * and a white noise generator. In addition to its sound producing capabilities,
 * the AY-3-8910 has two 8-bit I/O ports and the AY-3-8912 has one.
 *
 * DA0-DA7 are multiplexed data and address lines, as used in the range of GI
 * processors that this IC was originally part of. The BC1 and BDIR pins are
 * used to control the IC as follows:
 *
 *   Function       BDIR   BC1
 *	
 *   Inactive         0     0
 *   Read             0     1
 *   Write            1     0
 *   Latch address    1     1
 *
 * In addition to using BC1 and BDIR, A8 should be taken high to enable the
 * chip. IOA0-IOA7 are the input/output pins (the AY-3-8910 has an additional
 * set marked IOB0-IOB7). The IC does not contain an internal oscillator - the
 * clock input to the CLK pin should fall between 1MHz-2MHz, so a typical Z80
 * system's clock would need to be divided.
 *
 * Registers
 * ---------
 * The AY-3-8910/8912 contains 16 internal registers as follows:
 *
 *   Register        Function                        Range
 *
 *    0              Channel A fine pitch            8-bit (0-255)
 *    1              Channel A course pitch          4-bit (0-15)
 *    2              Channel B fine pitch            8-bit (0-255)
 *    3              Channel B course pitch          4-bit (0-15)
 *	  4              Channel C fine pitch            8-bit (0-255)
 *	  5              Channel C course pitch          4-bit (0-15)
 *	  6              Noise pitch                     5-bit (0-31)
 *	  7              Mixer                           8-bit (see below)
 *	  8              Channel A volume                4-bit (0-15, see below)
 *	  9              Channel B volume                4-bit (0-15, see below)
 *	 10              Channel C volume                4-bit (0-15, see below)
 *	 11              Envelope fine duration          8-bit (0-255)
 *	 12              Envelope course duration        8-bit (0-255)
 *	 13              Envelope shape                  4-bit (0-15)
 *	 14              I/O port A                      8-bit (0-255)
 *	 15              I/O port B                      8-bit (0-255)
 *
 * Notes:
 *
 *   - The AY-3-8912 does not contain register 15.
 *   - The volume registers (8, 9 and 10) contain a 4-bit setting but if bit 5
 *     is set then that channel uses the envelope defined by register 13 and
 *     ignores its volume setting.
 *   - The mixer (register 7) is made up of the following bits (low=enabled):
 *
 *     Bit: 7       6       5       4       3       2       1       0
 *          I/O     I/O     Noise   Noise   Noise   Tone    Tone    Tone
 *          B       A       C       B       A       C       B       A
 *
 *     The AY-3-8912 ignores bit 7 of this register.
 *
 * Envelopes
 * ---------
 * The AY-3-8910/8912 contains the following preset envelopes or waveforms (set
 * using control register 13). Note that these affect volume only and not the
 * pitch:
 *
 *    0      \__________     single decay then off
 *
 *    4      /|_________     single attack then off
 *
 *    8      \|\|\|\|\|\     repeated decay
 *
 *    9      \__________     single decay then off
 *
 *   10      \/\/\/\/\/\     repeated decay-attack
 *	           _________
 *   11      \|              single decay then hold
 *
 *   12      /|/|/|/|/|/     repeated attack
 *	          __________
 *   13      /               single attack then hold
 *
 *   14      /\/\/\/\/\/     repeated attack-decay
 *
 *   15      /|_________     single attack then off
 *
 * Pitch values
 * ------------
 * The course and fine pitch registers for each channel are used in the
 * following fashion (assuming channel A):
 *
 * Registers 0 and 1 operate together to form channel A's final pitch. The
 * eight least significant bits are sent to register 0 and the four most
 * significant bits are sent to register 1. The output frequency is equal to
 * the IC's incoming clock frequency divided by 16 and then further divided by
 * the number written to the course and fine pitch registers, so the higher the
 * number written to these, the lower the pitch. For example, if a frequency of
 * 1KHz was required and the IC's clock frequency was 1MHz, a total division
 * rate of 1000 would be needed. The sound generator itself divides by 16 so the
 * course and fine pitch registers must provide a further division by 62.5 (due
 * to the fact that 1000/16 is 62.5). A division rate of 62 or 63 will be
 * accurate enough, since the registers can only store whole numbers.
 * Therefore, 62 or 63 would be written to register 0 and 0 would be written to
 * register 1.
 *
 * The following frequency table may be useful in musical applications:
 *
 *   Note          Frequency (Hz)        Note          Frequency (Hz)
 *
 *   A             220                   D#            311.1
 *   A#            233.3                 E             329.63
 *   B             246.94                F             349.23
 *   middle C      261.63                F#            370
 *   C#            277.2                 G             392
 *   D             293.66                G#            415.3
 *******************************************************************************/

/********************************************************************************
 * ayPlay() - The main emulation of the AY sound chip.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
ayPlay(void)
{
	int incr0 = 0, incr1 = 0, incr2 = 0, increnv = 0, incrnoise = 0;
	int statenoise = 0, noisegen = 1;
	int counter0 = 0, counter1 = 0, counter2 = 0, countenv = 0, countnoise = 0;
	int r = 0, v = 0, i;
	int vol0, vol1, vol2, volnoise, envelope = 15;
	int c0, c1, l0, l1, l2;
	unsigned char buf[2];
	unsigned char sndbuf[SOUND_BUFSIZE];
	unsigned char PSG[16] = { 0, 0, 0, 0, 0, 0, 0, 255, 8, 8, 8, 0, 0, 0, 0, 0 };
	pid_t parent;

	parent = getppid();
	
	for (;;)
	{
		if (parent != getppid())
		{
			XzxMesg(XZX_WARN, "AY-3-8912 daemon: parent died");
			Quit(1);
		}
		
		/* read from pipe */
		while ((i = read(ayPipe[0], buf, v < 0 ? 1 : 2)) != 0)
		{
			if (i == -1)		/* Read error */
			{
				if (errno != EWOULDBLOCK)
				{
					XzxMesg(XZX_PERR, "AY-3-8912 daemon: can't read from pipe");
				}
				break;
			}
			if (v == -1)		/* Value expected */
			{
				v = buf[0];
			}
			else				/* Register expected */
			{
				r = buf[0];
				if (i == 1)
				{
					v = -1;		/* Next time read the value */
					break;
				}
				v = buf[1];
			}

#ifdef DEBUG
			if (GETCFG(debug) & D_AUDIO)
			{
				XzxMesg(XZX_DEBUG,
						"AY-3-8912 daemon: reg = %02x val = %02x", r, v);
			}
#endif				

			PSG[r] = v; ;		/* Store PSG registers new value */
		
			switch (r)
			{
				case 1:
					PSG[1] &= 0x0f;
				case 0:
					i = PSG[0] + (PSG[1] << 8);
					incr0 = i ? 1024 * PSG_CLOCK / aySampleRate * 4 / i : 0;
					break;
				case 3:
					PSG[3] &= 0x0f;
				case 2:
					i = PSG[2] + (PSG[3] << 8);
					incr1 = i ? 1024 * PSG_CLOCK / aySampleRate * 4 / i : 0;
					break;
				case 5:
					PSG[5] &= 0x0f;
				case 4:
					i = PSG[4] + (PSG[5] << 8);
					incr2 = i ? 1024 * PSG_CLOCK / aySampleRate * 4 / i : 0;
					break;
				case 6:
					i = PSG[6] & 0x1f;
					incrnoise = 1024 * PSG_CLOCK / aySampleRate / ( i ? i : 1 );
					break;
				case 8:
				case 9:
				case 10:
					PSG[r] &= 0x1f;
					break;
				case 11:
				case 12:
					i = PSG[11] + (PSG[12] << 8);
					increnv = i ? 1024 * PSG_CLOCK / aySampleRate * 4 / i *
						          SOUND_BUFSIZE : 0;
					countenv=0;
				break;
			}
		}
		
		envelope = envelopeForms[PSG[13]][(countenv >> 16) & 0x1f];
		if ((countenv += increnv) & 0xffe00000)
		{
			switch (PSG[13])
			{
				case 8:
				case 10:
				case 12:
				case 14:
					countenv -= 0x200000;
					break;
				default:
					countenv = 0x100000;
					increnv = 0;
			}
		}

		vol0 = PSG[ 8] < 16 ? PSG[ 8] : envelope;
		vol1 = PSG[ 9] < 16 ? PSG[ 9] : envelope;
		vol2 = PSG[10] < 16 ? PSG[10] : envelope;

		volnoise = ((PSG[7] & 010 ? 0 : vol0) +
					(PSG[7] & 020 ? 0 : vol1) +
					(PSG[7] & 040 ? 0 : vol2)) / 2;

		vol0 = PSG[7] & 0x01 ? 0 : vol0;
		vol1 = PSG[7] & 0x02 ? 0 : vol1;
		vol2 = PSG[7] & 0x04 ? 0 : vol2;

		for (i = 0; i < SOUND_BUFSIZE; i++)
		{
			/* These strange tricks are needed for getting rid of nasty
			   interferences between sampling frequency and "rectangular sound"
			   (which is also the output of real AY-3-8912) we produce. */

			c0 = counter0;
			c1 = counter0 + incr0;
			l0 = c0 & 0x8000 ? -16 : 16;
			if ((c0 ^ c1) & 0x8000)
			{
				l0 = l0 * (0x8000 - (c0 & 0x7fff) - (c1 & 0x7fff)) / incr0;
			}
			counter0 = c1 & 0xffff;
			
			c0 = counter1;
			c1 = counter1 + incr1;
			l1 = c0 & 0x8000 ? -16 : 16;
			if ((c0 ^ c1) & 0x8000)
			{
				l1 = l1 * (0x8000 - (c0 & 0x7fff) - (c1 & 0x7fff)) / incr1;
			}
			counter1 = c1 & 0xffff;
			
			c0 = counter2;
			c1 = counter2 + incr2;
			l2 = c0 & 0x8000 ? -16 : 16;
			if ((c0 ^ c1) & 0x8000)
			{
				l2 = l2 * (0x8000 - (c0 & 0x7FFF) - (c1 & 0x7fff)) / incr2;
			}
			counter2 = c1 & 0xffff;

			countnoise &= 0xffff;
			if ((countnoise += incrnoise) & 0xffff0000)
			{
				/* The following code is a random bit generator. */
				statenoise = ((noisegen <<= 1) & 0x80000000 ?
							  noisegen ^= 0x00040001 : noisegen ) & 1;
			}

			sndbuf[i] = 128 + ((l0 * vol0 + l1 * vol1 + l2 * vol2) / 16 +
								 (statenoise ? volnoise : -volnoise));
		}

		if (ayPaused)
		{
			sleep(1);
		}
		else
		{
#if defined(__linux__) || defined(XXX)
		
			/* We'll block here until next DMA buffer becomes free. It happens
			   once per (1<<SOUND_BUFSIZE_BITS)/aySampleRate seconds. */

			write(ayFd, sndbuf, SOUND_BUFSIZE);
#else
			/* Flush output first, don't care about return status. After this
			   write next buffer of audio data. This method produces a horrible
			   click on each buffer. */

			/* ioctl(ayFd, AUDIO_DRAIN ); */
			write(ayFd, sndbuf, SOUND_BUFSIZE );

#endif /* __linux__ || defined(XXX) */
		}
	}
}

/* EOF *************************************************************************/
