/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 *******************************************************************************/

#if defined(SUN_AUDIO) || defined(PCSPKR_AUDIO) || defined(XBELL_AUDIO)

static char speaker_c[] = "$Id: speaker.c,v 3.1 1996/10/03 19:02:23 erik Exp $";

/********************************************************************************
 * speaker.c - Sound implementations.
 * 
 * Currently we have:
 *   Speaker emulation via
 *     o PC Speaker    (Linux)
 *     o /dev/audio    (Sun Sparc, Linux, NEC 4800/330)

 *     o /dev/dsp      (Linux w/USS)
 *   BASIC BEEP emulation via:
 *     o X server bell (Linux)
 *******************************************************************************/

#ifdef SUN_AUDIO
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sun/audioio.h>
#endif

#ifdef PCSPKR_AUDIO
#include <unistd.h>
#include <asm/io.h>
#endif

#ifdef XBELL_AUDIO
#include <X11/Xlib.h>
#include "screen.h"
#endif

#include "z80.h"
#include "debug.h"
#include "io.h"
#include "util.h"
#include "speaker.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

/********************************************************************************
 * Forwards
 *******************************************************************************/

#ifdef SUN_AUDIO
static void speakerQuit(void);
static int speakerOpenDevice(void);
#endif

/********************************************************************************
 * Globals
 *******************************************************************************/

/********************************************************************************
 * Locals
 *******************************************************************************/

#ifdef SUN_AUDIO
static int spOn = 0;
static int spActive = 0;
static int spFd = -1;
static char *spDevice;
#endif

#ifdef PCSPKR_AUDIO
static int spOn = 0;
static int spActive = 0;
#endif

/********************************************************************************
 * SpeakerInit() - Initialize the speaker audio emulation.
 *
 * Input:  audioDev - the audio device to use
 *         audioOn - turn speaker on
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

#ifdef SUN_AUDIO

int
SpeakerInit(char *audioDev, int audioOn)
{
	OnQuit(speakerQuit);

	spDevice = audioDev;

	if (speakerOpenDevice() == -1)
	{
		return -1;
	}

	spActive = True;
	spOn = 1;

	if (!audioOn)
	{
		SpeakerOnOff();
	}
	
	return 0;
}

#endif /* SUN_AUDIO */

#ifdef PCSPKR_AUDIO

int
SpeakerInit(int audioOn)
{
	/* Linux PC speaker audio implementation. In theory, this ought to work on
	   any Unix running on a PC console, but I haven't tried it.
	   For this to succeed, it must be run with root permissions. Therefore we
	   do a setuid() and getuid() to give up root permissions once we know we've
	   succeeded. */

	if (ioperm(0x61, 1, 1))
	{
		XzxMesg(XZX_PWARN, "couldn't get permission for the speaker port");
		return -1;
	}

	setgid(getgid());
	setuid(getuid());

	spActive = True;
	spOn = 1;

	if (!audioOn)
	{
		SpeakerOnOff();
	}
	
	return 0;
}

#endif /* PCSPKR_AUDIO */

/********************************************************************************
 * speakerQuit() - Quit the speaker emulation.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

#ifdef SUN_AUDIO

static void
speakerQuit(void)
{
	if (spFd != -1)
	{
		close(spFd);
	}
}

#endif /* SUN_AUDIO */

/********************************************************************************
 * speakerOpenDevice() - Open the audio device for the speaker emulation.
 *
 * Input:  -
 *
 * Return:  0 - o.k.
 *         -1 - error
 *******************************************************************************/

#ifdef SUN_AUDIO

static int
speakerOpenDevice(void)
{
	/* Use "audioDev" (by default /dev/audio) to emulate the speaker. */
	if ((spFd = open(spDevice, O_WRONLY | O_NONBLOCK)) == -1)
	{
		XzxMesg(XZX_PWARN,
				"couldn't open audio device <%s> for writing", spDevice);
		return -1;
	}

	return 0;
}

#endif /* SUN_AUDIO */

/********************************************************************************
 * SpeakerOnOff - Turns speaker sound on/off.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

#if defined(SUN_AUDIO) || defined(PCSPKR_AUDIO)

void
SpeakerOnOff(void)
{
	if (spActive)
	{

		spOn = !spOn;

#ifdef SUN_AUDIO
		
		if (spOn)
		{
			speakerOpenDevice();
		}
		else
		{
			close(spFd);
		}

#endif /* SUN_AUDIO */
	}
}

#endif /* SUN_AUDIO || PCSPKR_AUDIO */

/********************************************************************************
 * SpeakerOutByte() - Output the sound bit on the ULA port.
 *
 * Input:  oldval - last value
 *         newval - new value
 *
 * Return: -
 *******************************************************************************/

#if defined(SUN_AUDIO) || defined(PCSPKR_AUDIO)

void
SpeakerOutByte(uns8 oldval, uns8 newval)
{
	if (spActive && spOn &&  ((newval ^ oldval) & B_SPEAKER))
	{
#ifdef SUN_AUDIO

		uns8 snd = newval & B_SPEAKER ? 0x1f : 0x00;
		/* ioctl(spFd, AUDIO_DRAIN); */
		write(spFd, &snd, sizeof(snd));

#endif /* SUN_AUDIO  */

#ifdef PCSPKR_AUDIO

		if (newval & B_SPEAKER)
		{
			outb(inb(0x61) | 0x02, 0x61);
		}
		else
		{
			outb(inb(0x61) & 0xfc, 0x61);
		}

#endif /* PCSPKR_AUDIO */
	}
}

#endif /* SUN_AUDIO || PCSPKR_AUDIO */

/********************************************************************************
 * RingBell() - Ring the X server bell at the specified frequency and duration.
 *
 * Input:  freq     - frequency
 *         duration - duration
 *
 * Return: -
 *******************************************************************************/

#ifdef XBELL_AUDIO

void
RingBell(int freq, int duration)
{
	XKeyboardState kbs, oldKbs;

#ifdef DEBUG
	if (GETCFG(debug) & D_AUDIO)
	{
		XzxMesg(XZX_DEBUG, "BEEP: freq = %d Hz, time = %d ms", freq, duration);
	}
#endif

	XGetKeyboardControl(Dpy, &oldKbs);

	kbs.bell_pitch = freq;
	kbs.bell_duration = duration;
	kbs.bell_percent = GETCFG(bellVol);

#define KEYBOARD_FLAG (KBBellPitch | KBBellDuration | KBBellPercent)
	
	XChangeKeyboardControl(Dpy, KEYBOARD_FLAG, (XKeyboardControl *)&kbs);
	XBell(Dpy, 0);
	XChangeKeyboardControl(Dpy, KEYBOARD_FLAG, (XKeyboardControl *)&oldKbs);

#undef KEYBOARD_FLAG
	
	XFlush(Dpy);
}

#endif /* XBELL_AUDIO */

#endif /* SUN_AUDIO || PCSPKR_AUDIO || XBELL_AUDIO */

/* EOF *************************************************************************/
