/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: emul.c,v 3.0 1996/10/03 10:23:45 erik Exp $";

/********************************************************************************
 * emul.c - This is the heart of the Spectrum emulation.
 *******************************************************************************/

#include <stdio.h>
#include <sys/types.h>

#include <OS.h>
#include <SupportDefs.h>

#include "ZXConfig.h"
#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "dis.h"
#include "fdc.h"
#include "if1.h"
#include "screen.h"
#include "emul.h"

#include "z80.c"				/* The Z80 emulation kernel */

/********************************************************************************
 * Defines
 *******************************************************************************/

/********************************************************************************
 * Forwards
 *******************************************************************************/

static void frameInterrupt(int);
#ifdef SLOWDOWN
static void nop(void);
#endif

/********************************************************************************
 * Globals
 *******************************************************************************/

int Vline = 0;					/* # of video line currently displayed */

/********************************************************************************
 * Locals
 *******************************************************************************/

static int32 interrupted = 0;		/* # of interrupts pending */
static sem_id interruptSem;		/* Semaphore for interrupt handling */
static int nInterrupts = 0;		/* # of interrupts since powerup */
static int refreshInterval;		/* # of interrupts until screen refresh */
static int cyclesPerLine;			/* # of T-states per frame line */

/********************************************************************************
 * frameInterrupt() - Called periodically by interval timer thread.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
frameInterrupt(void)
{
	if (atomic_add(&interrupted, 1) == 0) {
		release_sem(interruptSem);
	}
}

/********************************************************************************
 * nop() - Do nothing.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

#ifdef SLOWDOWN

static void
nop(void)
{
}

#endif

/********************************************************************************
 * Emulate() - Main emulator loop.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
Emulate(void)
{
	register uns8 op = 0;
#ifdef SLOWDOWN
	int i;
#endif

	BasicRom = GET_BASIC_ROM(GETCFG(machine));
	cyclesPerLine = GETCFG(cycles) / 312;
	refreshInterval = GETCFG(refreshRate);

	/* Start interrupts */
	interruptSem = create_sem(0, "Frame interrupt");
	if (interruptSem < B_OK) {
		return;
	}
	IntFrequency(1000 * GETCFG(ms), frameInterrupt);
	
	for (;;) {
		/* The 50 Hz interrupt is synchronized with the video signal generation
		   by the ULA; both the interrupt and the video signal are generated by
		   it. Many programs use the interrupt to synchronize with the frame
		   cycle. Some use it to generate fantastic effects, such as full-screen
		   characters, full-screen horizon (Aquaplane) or pixel colour (Uridium
		   for instance). */

		if (Tstates >= cyclesPerLine)
		{
			Tstates -= cyclesPerLine;

			if (!Vline)
			{
				nInterrupts++;
				CheckEvents();

				/* Every 25 interrupts -> twice/sec. */
				if (!(nInterrupts % 25))
				{
					DoFlashing();
				}

				if (!(nInterrupts % refreshInterval))
				{
					ScreenRefresh();
				}

				while (acquire_sem(interruptSem) == B_INTERRUPTED)
					;
				interrupted = 0;

				Z80_INTERRUPT
			}
			Vline = ++Vline % 312;
		}

#ifdef SLOWDOWN
		for (i = 0; i < GETCFG(slowdown); i++)
		{
			nop();
		}
#endif

		/* ROM2 on the +3 contains all the DOS routines */
		if (GETCFG(machine) == 3 && RPAGE(0) == ROM2)
		{
			DdCheck();
		}

		if ((PC == 0x0008 || PC == 0x1708) && RPAGE(0) == BasicRom &&
			GETCFG(if1_active))
		{
			PageIn(0, ROM2);
		}

		if (PC == 0x0066 && RPAGE(0) == BasicRom && GETCFG(mf128_active))
		{
			PageIn(0, ROM3);
		}

#ifdef DEBUG
		if (GETCFG(debug) & (D_CPU | D_SYSVAR | D_MNEMO))
		{
			DebugZ80();
		}
#endif

		Z80_M1(op)

		if (PC == 0x0701 && RPAGE(0) == ROM2 && GETCFG(if1_active))
		{
			PageIn(0, BasicRom);
		}

		switch (op)
		{
#			include "z80_ops.c"
		}

		Z80_DELAYED_EI
	}
}

/* EOF *************************************************************************/
