/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Martin Smith (msmith@lssec.bt.co.uk)
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: if1.c,v 3.1 1996/10/03 19:06:40 erik Exp $";

/********************************************************************************
 * An emulation of the ZX INTERFACE 1 MICRODRIVES
 *
 * By Martin H. Smith
 *
 * This code has nothing to do with my employers, they certainly don't pay
 * me to enjoy myself like this! If only!
 *
 * Saturday 12th March 1994
 *
 * Note: This emulation has been produced from my own knowledge and guesses
 * at how the Interface 1 hardware works based on a full disassembly I
 * jointly wrote 10 years or so ago. This means I cannot vouch that the port
 * operations are emulated exactly as any other Spectrum microdrive emulator
 * does but the .MDR files should be compatible with other emulators.
 *******************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "ZXConfig.h"
#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "util.h"
#include "if1.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

enum { FromTty, FromFile, FromPipe };	/* where RS232 input can come from */

/********************************************************************************
 * Forwards
 *******************************************************************************/

static void setBreakpoints(void);
static void setBlocking(int, int);
static int classifyDescriptor(int);
static int testCartridgeFile(int);
static void setupCartridgeFiles(char *[MAX_DRIVES]);
static void if1Quit(void);
static int loadCartridgeFile(int);
static void unloadCartridgeFile(int);
static void flushCache(void);
static void startDrive(int);
static void stopDrive(int);

/********************************************************************************
 * Globals
 *******************************************************************************/

/********************************************************************************
 * Locals
 *******************************************************************************/

static int stdinFileType, stdoutFileType;

static long	mdriveBytesOut = 0L;		/* Statistics */
static long mdriveBytesIn = 0L;

static int gapClock = 0;				/* Counter for GAP bit */
static int gapBit = 0;					/* State of GAP bit */

static mdriveData drives[MAX_DRIVES + 1];

static long	driveSeq = 0L;				/* Cache sequence number */

static int preamble = 0;				/* Preamble skip byte counter */

static int driveRunning = 0;			/* Number of current drive */
static int driveCount = 0;				/* Drive count for start / stop */

/********************************************************************************
 * setBreakpoints() - Put down our breakpoints, if the appropriate rom has been
 *                    loaded into memory.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
setBreakpoints(void)
{
	if (RealMemory[ROM2][RS232_IP_ADDRESS_V1] == RS232_IP_BPT_BYTE &&
		RealMemory[ROM2][RS232_OP_ADDRESS_V1] == RS232_OP_BPT_BYTE)
	{
		RealMemory[ROM2][RS232_IP_ADDRESS_V1] = 0xed;
		RealMemory[ROM2][RS232_IP_ADDRESS_V1 + 1] = RS232_IP_BPT;
		RealMemory[ROM2][RS232_OP_ADDRESS_V1] = 0xed;
		RealMemory[ROM2][RS232_OP_ADDRESS_V1 + 1] = RS232_OP_BPT;
	}
	else if (RealMemory[ROM2][RS232_IP_ADDRESS_V2] == RS232_IP_BPT_BYTE &&
			 RealMemory[ROM2][RS232_OP_ADDRESS_V2] == RS232_OP_BPT_BYTE)
	{
		RealMemory[ROM2][RS232_IP_ADDRESS_V2] = 0xed;
		RealMemory[ROM2][RS232_IP_ADDRESS_V2 + 1] = RS232_IP_BPT;
		RealMemory[ROM2][RS232_OP_ADDRESS_V2] = 0xed;
		RealMemory[ROM2][RS232_OP_ADDRESS_V2 + 1] = RS232_OP_BPT;
	} 
	else
	{
		XzxMesg(XZX_WARN, "unknown IF1 rom, RS232 to stdio is disabled");
	}
}

/********************************************************************************
 * setBlocking() -  Make file descriptor <fd> either blocking or non-blocking
 *                  dependng on mode.
 *
 * Input:  fd   - file descriptor
 *         mode - blocking mode
 *                  0 = BLOCKING
 *                  1 = NONBLOCKING
 *
 * Return: -
 *******************************************************************************/

static void
setBlocking(int fd, int mode)
{
	int status;
	
	if  ((status = fcntl(fd, F_GETFL)) != -1)
	{

#ifdef O_NONBLOCK
		status = mode ? status | O_NONBLOCK : status & ~O_NONBLOCK;
#else
		status = mode ?	status | O_NDELAY : status & ~O_NDELAY;
#endif

		if (fcntl(fd, F_SETFL, status) != -1)
		{
			return;
		}
	}
	
	XzxMesg(XZX_PERR, "couldn't make fd %d %sblocking", fd, mode ? "non" : "");
}

/********************************************************************************
 * classifyDescriptor() - Decide whether file descriptor <fd> is a normal file,
 *                        a tty, or a pipe. This affects the semantics of reads
 *                        and writes via RS232.
 *
 * Input:  fd - file descriptor
 *
 * Return: FromTty  = tty
 *         FromFile = file
 *         FromPipe = pipe
 *         -1       = error
 *******************************************************************************/

static int
classifyDescriptor(int fd)
{
	struct stat buf;

	if (isatty(fd))
	{
		return FromTty;
	}

	fstat(fd, &buf);

	if (S_ISREG(buf.st_mode))
	{
		return FromFile;
	}
	else if (S_ISFIFO(buf.st_mode))
	{
		return FromPipe;
	}
	else
	{
		XzxMesg(XZX_WARN, "don't know where fd %d is coming from", fd);
		return -1;
	}
}

/********************************************************************************
 * testCartridgeFile() - Ensure that a cartridge file is a normal file and has
 *                       the correct size.
 *
 * Input:  drive - # of microdrive (1-8)
 *
 * Return: 1 - o.k.
 *         0 - error
 *******************************************************************************/

static int
testCartridgeFile(int drive)
{
	struct stat sbuff;

	if (!fstat(drives[drive].fd, &sbuff)) 	/* Stat ok */
	{
		if (S_ISREG(sbuff.st_mode) && sbuff.st_size == CART_FILE_SIZE)
		{
			return 1;						/* Good file size */
		}
		else
		{
			XzxMesg(XZX_ERR, "cart file %s in drive %d is bad",
					drives[drive].filename, drive);
		}
	}

	return 0;								/* Bad file */
}

/********************************************************************************
 * setupCartridgeFiles() - Try and load the cartridge files into the drives.
 *                         Set up the initial arrays and handle write
 *                         protection
 *
 * Input:  cname - array with microdrive file names
 *
 * Return: -
 *******************************************************************************/

static void
setupCartridgeFiles(char *cname[MAX_DRIVES])
{
	int f;

	for (f = 1; f <= MAX_DRIVES; ++f)
	{
		drives[f].sequence = 0L;			/* Cache sequence */
		drives[f].protect = 1;				/* Protected */
		drives[f].tapepos = 0L;				/* At start */
		drives[f].loaded = 0;				/* No cartridge */
		drives[f].filename = cname[f - 1];	/* Cart file */
		drives[f].buffer = NULL;
		drives[f].fd = -1;

		if (drives[f].filename[0])			/* not null filename */
		{
			if ((drives[f].fd = open(drives[f].filename, O_RDONLY)) > 0)
			{
				if (testCartridgeFile(f))
				{
					drives[f].loaded = 1;	/* a valid file */
				}
				close(drives[f].fd);
				drives[f].fd = -1;
			}
		}
	}
}

/********************************************************************************
 * if1Quit() - Close down the IF1, all drives should be stopped at this point
 *             anyway but it writes back cart files in case the emulator was
 *             stopped part way through an operation by a quit / reset etc.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
if1Quit(void)
{
	int f;
	
	if (GETCFG(if1_active))
	{
		for (f = 1; f <= MAX_DRIVES; ++f)
		{
			if (drives[f].loaded && drives[f].buffer)
			{
				unloadCartridgeFile(f);
			}
		}

#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			/* Tell the user how many bytes were written */
			XzxMesg(XZX_DEBUG, "%d bytes out to microdrive, %d bytes in",
					mdriveBytesOut, mdriveBytesIn);
		}
		
#endif

		setBlocking(STDIN_FILENO, 0);
		setBlocking(STDOUT_FILENO, 0);
	}
}

/********************************************************************************
 * If1Init() - Set up our interface emulation. If the user wants and IF1 and we
 *             manage to load our ROM image file then activate the emulation
 *             and load the cartridge files.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
If1Init(void)
{
	if (!GETCFG(if1_active))
	{
		return;
	}
	
	OnQuit(if1Quit);

	setBreakpoints();

	setupCartridgeFiles(GETCFG(cartfiles));

	setBlocking(STDIN_FILENO, 1);
	stdinFileType = classifyDescriptor(STDIN_FILENO);

	setBlocking(STDOUT_FILENO, 1);
	stdoutFileType = classifyDescriptor(STDOUT_FILENO);

#ifdef DEBUG
	if (GETCFG(debug) & D_IF1)
	{
		XzxMesg(XZX_DEBUG, "IF1 initialised successfully");
	}
#endif
}

/********************************************************************************
 * loadCartridgeFile() - Load up a cartridge file.
 *
 * Input:  d - # of microdrive (1-8)
 *
 * Return: 1 - o.k.
 *         0 - error
 *******************************************************************************/
 
static int
loadCartridgeFile(int d)
{
	int res = 0;

	drives[d].fd = open(drives[d].filename, O_RDONLY);

	drives[d].tapepos /= (long)CART_FILE_SECTOR_SIZE;
	drives[d].tapepos *= (long)CART_FILE_SECTOR_SIZE;
	drives[d].loaded = 0;
	drives[d].dirty = 0;
	drives[d].protect = 0;
	
	if (drives[d].fd <= 0)
	{
		XzxMesg(XZX_PERR, "can't open cartidge file %s", drives[d].filename);
		return 0;
	}
	
	if (!(drives[d].buffer = (uns8 *) malloc(CART_FILE_SIZE)))
	{
		XzxMesg(XZX_FATAL, "malloc failed in loadCartridgeFile()");
	}

	if (read(drives[d].fd, drives[d].buffer, CART_FILE_SIZE))
	{
		drives[d].loaded = 1;

		/* The flag may say writeable ... */
		drives[d].protect = drives[d].buffer[CART_FILE_SIZE - 1] ? 1 : 0;

		/* But the file might be read only */
		if (access(drives[d].filename, W_OK) != 0)
		{
			drives[d].protect = 1;
		}

#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "read %s cart file %d [%s]",
					drives[d].protect ? "R/O" : "R/W", d, drives[d].filename);
		}
#endif

		res = 1;
	}
	else
	{
		XzxMesg(XZX_PERR, "can't read cartridge file %s", drives[d].filename);

		free(drives[d].buffer);
		drives[d].buffer = NULL;
	}

	close(drives[d].fd);
	drives[d].fd = -1;
	
	return res;
}

/********************************************************************************
 * unloadCartridgeFile() - Get rid of a cartridge file, writing it back if
 *                         necessary.
 *
 * Input:  d - # of microdrive (1-8)
 *
 * Return: -
 *******************************************************************************/

static void
unloadCartridgeFile(int d)
{
	if (drives[d].dirty && !drives[d].protect)
	{
		drives[d].fd = open(drives[d].filename, O_RDWR);

		if (drives[d].fd > 0) 					  /* File is open */
		{
			if (write(drives[d].fd, drives[d].buffer, CART_FILE_SIZE)
				!= CART_FILE_SIZE)
			{
				XzxMesg(XZX_PERR, "failed to write cart file %s",
						drives[d].filename);
			}
#ifdef DEBUG
			else
			{
				if (GETCFG(debug) & D_IF1)
				{
					XzxMesg(XZX_DEBUG, "wrote cart file %d [%s]",
							d, drives[d].filename);
				}
				
			}
#endif
			close(drives[d].fd);
			drives[d].fd = -1;
		}
		else
		{
			XzxMesg(XZX_PERR, "failed to open cart file %s for write",
					drives[d].filename);
		}
	}
#ifdef DEBUG
	else
	{
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "unloaded cartridge %d", d);
		}
	}
#endif
	
	drives[d].dirty = 0;
	drives[d].loaded = 0;
}

/********************************************************************************
 * flushCache() - We are allowed to keep up to a certain number of drive images
 *                in memory. if we are at that number then free up the drive
 *                which was used the longest time ago.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
flushCache(void)
{
	int count = 0;
	int picked = 0;
	long sequence = 0L;
	int f;

	for (f = 1; f <= MAX_DRIVES; ++f)
	{
		if (drives[f].buffer) 						/* This drive in RAM */
		{
			++count;

			/* find the oldest drive */
			if (sequence == 0 || drives[f].sequence < sequence)
			{
				sequence = drives[f].sequence;
				picked = f;
			}
		}
	}

	if (picked && count >= KEEP_DRIVES_IN_RAM)  	/* Need to flush */
	{
		free(drives[picked].buffer);
		drives[picked].buffer = 0;

#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "threw drive %d out of RAM", picked);
		}
#endif
	}
}

/********************************************************************************
 * startDrive() - Start the 'motor' of a microdrive.
 *
 * Input:  which - # of microdrive (1-8)
 *
 * Return: -
 *******************************************************************************/

static void
startDrive(int which)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IF1)
	{
		XzxMesg(XZX_DEBUG, "drive %d on", which);
	}
#endif

	gapClock = 0;
	gapBit = 0;

	drives[which].tapepos /= (long) CART_FILE_SECTOR_SIZE;
	drives[which].tapepos *= (long) CART_FILE_SECTOR_SIZE;

	if (driveRunning && driveRunning != which)
	{										/* Turn this one off */
		stopDrive(driveRunning);
		driveRunning = 0;
	}
		
	if (driveRunning != which)				/* This one not going */
	{
		if (drives[which].buffer)
		{									/* Already in memory */
			driveRunning = which;
			drives[driveRunning].sequence = ++driveSeq;
		}
		else
		{									/* Clear out cache */
			flushCache();
			
			if (loadCartridgeFile(which))
			{
				drives[driveRunning].sequence = ++driveSeq;
				driveRunning = which;
			}
#ifdef DEBUG
			else
			{
				if (GETCFG(debug) & D_IF1)
				{
					XzxMesg(XZX_DEBUG, "but no cartridge");
				}
			}
#endif
		}
	}
}

/********************************************************************************
 * stopDrive() - Stop a microdrive motor.
 *
 * Input:  d - # of micrdrive (1-8)
 *
 * Return: -
 *******************************************************************************/

static void
stopDrive(int d)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IF1)
	{
		XzxMesg(XZX_DEBUG, "drive %d stopped", d);
	}
#endif
	
	/* See if we are stopping the current one */
	if (d == driveRunning)
	{
		unloadCartridgeFile(d);
		driveRunning = 0;
	}
}

/********************************************************************************
 * OutPortEF() - Output a byte to the control port of the microdrives (0xef).
 *               Note this is not a perfect emulation, in particular I have
 *               not emulated the behaviour of outing 0 erasing cartridges!
 *
 * Input:  byte - data byte
 *
 * Return: -
 *******************************************************************************/

void
OutPortEF(uns8 byte)
{
	static int lastByte = 0;

	if (driveCount > 7 || driveCount < 0)
	{
		driveCount = 0;  /* Insanity check on drive count */
	}

	/* If we're about to start writing account for 12 bytes which must be
	   skipped (the preamble) before the actual data */
	if (byte == 0xe2 && !preamble)
	{
		preamble = 12;
	}

	/* Step the count forwards, turn things on and off */
	if (byte == 0xee && lastByte == 0xee)
	{
		driveCount = 0;  /* Reset count */
		byte = 0;

#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "interface 1 drive_count reset");
		}
#endif
	}
	else if (byte == 0xec && lastByte == 0xee)
	{ 					 /* Turn on drive */
		++driveCount;
		startDrive(9 - driveCount);
	}
	else if (byte == 0xed && lastByte == 0xef)
	{ 					 /* Turn off drive */
		++driveCount;
		stopDrive(9 - driveCount);
	}

	lastByte = byte;
}

/********************************************************************************
 * OutPortE7() - Send a byte to the microdrive data port. The byte is thrown
 *               away if a drive is not running.
 *
 * Input:  byte - data byte
 *
 * Return: -
 *******************************************************************************/

void
OutPortE7(uns8 byte)
{
	++mdriveBytesOut;  /* For stats freaks! */

	if (driveRunning && !preamble)  /* A byte to write out */
	{
		/* Find offset in current sector */
		int pos = drives[driveRunning].tapepos % CART_FILE_SECTOR_SIZE;

		/* If we lose write sync it is very serious! This is detected by
		   writing a sector header with a reset flag bit or record header with
		   a set flag bit. If this happens we mark the cartridge as write
		   protected. This terminates the operation with an error, without
		   writing back the duff cart file */

		if ((!pos && (byte & 1) != 1) || (pos == 15 && (byte & 1) != 0))
		{
			XzxMesg(XZX_ERR, "microdrive write sync lost %d",
					drives[driveRunning].tapepos);
			drives[driveRunning].protect = 1;  /* Don't write the corruption */
		}

		/* Store the byte away */
		drives[driveRunning].buffer[drives[driveRunning].tapepos] = byte;
		drives[driveRunning].dirty = 1;  /* Cartridge written to */

		/* 3 troublesome bytes are written after the ends of headers and
		   sectors. Lucky I remembered this one! */
		
		if (!preamble && (pos == (CART_FILE_SECTOR_SIZE - 1) || pos == 14))
		{
			preamble = 3;
		}

		/* Has the 'tape' wrapped round ? */
		if (++(drives[driveRunning].tapepos) == CART_FILE_SIZE - 1)
		{
			drives[driveRunning].tapepos = 0L;

#ifdef DEBUG
			if (GETCFG(debug) & D_IF1)
			{
				XzxMesg(XZX_DEBUG, "drive %d revolution (write)",
						driveRunning);
			}
#endif

		}
	}
	else if (driveRunning && preamble > 0)
	{
		--preamble;	 /* Ignore a preamble byte */
	}
}

/********************************************************************************
 * InPortE7() - Read in from the microdrive data port, if the drive is running
 *              we return the next byte of the tape loop. If not we don't
 *              emulate the IF1 behaviour of locking up. I hope no one minds.
 *
 *              Note there is no read sync lost check as for write as
 *              unformatted carts may be placed in drives.
 *
 * Input:  -
 *
 * Return: input data
 *******************************************************************************/

uns8
InPortE7(void)
{
	int res = 0xff;

	if (driveRunning)
	{
		++mdriveBytesIn;

		res = drives[driveRunning].buffer[drives[driveRunning].tapepos];
		
		if (++drives[driveRunning].tapepos == CART_FILE_SIZE - 1)
		{
			drives[driveRunning].tapepos = 0L;

#ifdef DEBUG
			if (GETCFG(debug) & D_IF1)
			{
				XzxMesg(XZX_DEBUG, "drive %d revolution (read)", driveRunning);
			}
#endif
		}
	}

	return res;
}

/********************************************************************************
 * InPortEF() - Read a byte of data from the microdrive control port. Does not
 *              do an exact emulation, just good enough to work and does not
 *              deal with bits not related to the microdrive.
 *
 *              If a drive is not running the status bits remain high and the
 *              interface ROM will time us out with a Microdrive not present
 *              error.
 *
 * Input:  -
 *
 * Return: input data
 *******************************************************************************/

uns8
InPortEF(void)
{
	uns8 res = 0xe0;  /* Basic result */

	if (driveRunning)
	{
		/* Get notional position within a sector */
		int pos = drives[driveRunning].tapepos % CART_FILE_SECTOR_SIZE;
		int syncBit = SYNC_BIT;

		if (pos == 0 || pos == 15)	/* Sync at end of preamble */
		{
			syncBit = 0;  /* Sync bit reset, ready to start I/O */
		}
		else
		{
			/* Feed the tape past the head until we reach a point where a sync
			   signal should occur */
			if ((++drives[driveRunning].tapepos) == CART_FILE_SIZE - 1)
			{
				drives[driveRunning].tapepos = 0L;

#ifdef DEBUG
				if (GETCFG(debug) & D_IF1)
				{
					XzxMesg(XZX_DEBUG, "drive %d revolution (sync)",
							driveRunning);
				}
#endif				
			}
		}

		if (!drives[driveRunning].protect)  /* Set up prot bit */
		{
			res |= WRITE_PROT_BIT;
		}

		/* Just do the GAP bit as a simple clock, at least for now */
		if ((++gapClock % GAP_CLOCK_RATE) == 0)
		{
			gapBit = !gapBit;
		}

		if (gapBit)
		{
			res |= GAP_BIT;
		}

		res |= syncBit; 
	}
	else
	{
		res |= GAP_BIT | SYNC_BIT | WRITE_PROT_BIT;  /* All high */
	}

	return res;
}

/********************************************************************************
 * RS232IpTrap() - RS232 Input, A register is set to next byte from stdin.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
RS232IpTrap(void)
{
	char c;
	int ret;

	if  ((ret = read(fileno(stdin), (void *)&c, 1)) == 1)
	{
#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "read code %d from stdin", (unsigned int)c);
		}
#endif

		A = (c == 10 && GETCFG(translate_nl)) ? 13 : c;
	}
	else
	{
		switch (stdinFileType)
		{
			case FromPipe:
			case FromTty:
				if (errno == 0 || errno == EAGAIN)
				{
					CLR(C_FLAG); /* No input ready */
					SET(Z_FLAG);
				}
				else
				{
					XzxMesg(XZX_PERR, "read error on stdin - got %d bytes", ret);
				}
				break;
			case FromFile:
				if (!ret)
				{
					CLR(C_FLAG); /* EOF */
					CLR(Z_FLAG);
				}
				else
				{
					XzxMesg(XZX_PERR, "read error on stdin - got %d bytes", ret);
				}
				break;
			default:
				XzxMesg(XZX_ERR, "don't know what I'm reading");
				break;
		}
	}
	
	RET();
}

/********************************************************************************
 * RS232OpTrap() - RS232 Output, Byte in A register is sent to stdout.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
RS232OpTrap(void)
{
	char c;
	
	if (A == 13 && GETCFG(strip_nl))
	{
		return;  /* Don't do anything */
	}

	c = (A == 13 && GETCFG(translate_nl)) ? 10 : A;

	if (write(fileno(stdout), (void *)&c, 1) == 1)
	{
		fflush(stdout);

#ifdef DEBUG
		if (GETCFG(debug) & D_IF1)
		{
			XzxMesg(XZX_DEBUG, "wrote code %d to stdout", (unsigned int)A);
		}
#endif
	}
	else
	{
		XzxMesg(XZX_PERR, "couldn't write to stdout");
	}
	
	RET();
}

/* EOF *************************************************************************/
