/********************************************************************************
 * Copyright (c) Erik Kunze  1995, 1996
 *               Jens Kilian 1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: io.c,v 3.1 1996/10/03 19:06:50 erik Exp $";

/********************************************************************************
 * io.c - Input & output port emulation.
 *******************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "ZXConfig.h"
#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "util.h"
#include "audio.h"
#include "joystick.h"
#include "emul.h"
#include "if1.h"
#include "screen.h"
#include "io.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

/********************************************************************************
 * Forwards
 *******************************************************************************/

static inline void outBankm128(uns8);
static inline void outBankmPl3(uns8);

/********************************************************************************
 * Globals
 *******************************************************************************/

/* For 128K and +3, remember the last values written to the bankm & bank678
   output ports.  Also, the screen selection flag; either 5 or 7. */

uns8 Last0x7FFD;				/* Memory paging on 128K */
uns8 Last0x1FFD;				/* Additional memory paging on +3 */
uns8 Last0xFFFD;				/* sound chip */

int ScreenSelect = RAM5;

uns8 KeyPorts[9];
uns8 InPorts[256];

int BasicRom;					/* In which ROM does BASIC live? */

int LastBorderColor = WHITE;	/* Dull white border */

uns8 PSG[16] = { 0, 0, 0, 0, 0, 0, 0, 255, 8, 8, 8, 0, 0, 0, 0, 0 };

/********************************************************************************
 * Locals
 *******************************************************************************/

static uns8 outPorts[256];

/********************************************************************************
 * The 128K and +3 memory management system.  We provide 2 different
 * implementations of port 0x7ffd, since it behaves slightly differently
 * between the 128K and +3 machines.
 *
 * When memory is being paged, interrupts should be disabled and the stack
 * should be in an area which is not going to change. If normal interrupt code
 * is to run, then the system variable at 5B5Ch (23388) must be kept updated
 * with the last value sent to port 7FFDh. It is not possible to read this port.
 *******************************************************************************/

/********************************************************************************
 * PAGING ON THE 128 AND +2
 *
 * On these computers, memory is entirely controlled by port 7FFDh. The byte to
 * output will be interpreted thus:
 *
 *   Bits 0-2: RAM page (0-7) to map into memory at 0C000h
 *   Bit    3: Select normal (0) or shadow (1) screen. The normal screen is in
 *             bank 5 and normally appears at 4000h; the shadow screen is in
 *             bank 7 and can be switched in at 0C000h
 *   Bit    4: ROM select. ROM 0 is the 128k editor and menu system; ROM 1
 *             contains 48k BASIC.
 *   Bit    5: If set, memory paging will be disabled and further output to this
 *             port will be ignored until the computer is reset.
 *
 * The memory map of these computers is:
 *
 * FFFFh+--------+--------+--------+--------+--------+--------+--------+--------+
 *      | Bank 0 | Bank 1 | Bank 2 | Bank 3 | Bank 4 | Bank 5 | Bank 6 | Bank 7 |
 *      |        |        |(also at|        |        |(also at|        |        |
 *      |        |        | 8000h) |        |        | 4000h) |        |        |
 *      |        |        |        |        |        | screen |        | screen |
 * C000h+--------+--------+--------+--------+--------+--------+--------+--------+
 *      | Bank 2 |          Any one of these pages may be switched in.
 *      |        |
 *      |        |
 *      |        |
 * 8000h+--------+
 *      | Bank 5 |
 *      |        |
 *      |        |
 *      | screen |
 * 4000h+--------+--------+
 *      | ROM 0  | ROM 1  | Either ROM may be switched in.
 *      |        |        |
 *      |        |        |
 *      |        |        |
 * 0000h+--------+--------+
 *
 * Memory banks 4-7 are contended, ie the processor shares them with the ULA.
 * This reduces the speed of memory access in these banks.
 *******************************************************************************/

/********************************************************************************
 * outBankm128() - Spectrum 128k memory management.
 *
 * Input:  val - control byte
 *
 * Return: -
 *******************************************************************************/

static inline void
outBankm128(uns8 val)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IO)
	{
		XzxMesg(XZX_DEBUG, 
				"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
				val & BIT_SELRAM, (val & BIT_SELROM) >> 4, (val & BIT_SELSCREEN) >> 3, 
				PAGING_DISABLED(Last0x7FFD) ? "dis" : "en");
	}
#endif

	if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}

	/* First bit D3, screen select; 0 selects page 5, 1 selects page 7 */
	ScreenSelect = val & BIT_SELSCREEN ? RAM7 : RAM5;
	if ((val & BIT_SELSCREEN) != (Last0x7FFD & BIT_SELSCREEN))
	{
		ForceScreenRefresh();
	}

	/* Now bit D5, the page disable bit */
	Last0x7FFD = val;

	/* Now check bits D0-D2; RAM select at 0xC000 */
	PageIn(3, RAM0 + (val & BIT_SELRAM));

	/* And bit D4, ROM select at 0x0000 */
	PageIn(0, ROM0 + ((val & BIT_SELROM) >> 4));
}

/********************************************************************************
 * PAGING ON THE +2A AND +3
 *
 * The basic principle is the same. However, the +2A and +3 have four ROMs
 * rather than two, and certain extra memory configurations.
 *
 * Port 7FFDh behaves in the same way as before, except that bit 4 is now the
 * low bit of the ROM selection. The extra features are controlled by
 * port 1FFDh. This port is also write-only, and its last value should be saved
 * at 5B67h (23399).
 *
 * Port 1FFDh responds thus:
 *
 *   Bit 0: Paging mode. 0=normal, 1=special
 *   Bit 1: In normal mode, ignored.
 *   Bit 2: In normal mode, high bit of ROM selection. The four ROMs are:
 *          ROM 0: 128k editor, menu system and self-test program
 *          ROM 1: 128k syntax checker
 *			ROM 2: +3DOS
 *			ROM 3: 48 BASIC
 *	 Bit 3: Disk motor; 1=on, 0=off
 *	 Bit 4: Printer port strobe.
 *
 * When special mode is selected, the memory map changes to one of four
 * configurations specified in bits 1 and 2 of port 1FFDh:
 *
 *       Bit 2 =0    Bit 2 =0    Bit 2 =1    Bit 2 =1
 *       Bit 1 =0    Bit 1 =1    Bit 1 =0    Bit 1 =1
 * FFFFh+--------+  +--------+  +--------+  +--------+
 *      | Bank 3 |  | Bank 7 |  | Bank 3 |  | Bank 3 |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 *		|        |  | screen |  |        |  |        |
 * C000h+--------+  +--------+  +--------+  +--------+
 *	    | Bank 2 |  | Bank 6 |  | Bank 6 |  | Bank 6 |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 * 8000h+--------+  +--------+  +--------+  +--------+
 *		| Bank 1 |  | Bank 5 |  | Bank 5 |  | Bank 7 |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 *		|        |  | screen |  | screen |  | screen |
 * 4000h+--------+  +--------+  +--------+  +--------+
 *		| Bank 0 |  | Bank 4 |  | Bank 4 |  | Bank 4 |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 *		|        |  |        |  |        |  |        |
 * 0000h+--------+  +--------+  +--------+  +--------+
 *
 * An example of a typical bank switch on the 128 is:
 *
 *     LD      A,(5B5Ch)       ;Previous value of port
 *     AND     0F8h
 *     OR      4               ;Select bank 4
 *     LD      BC,7FFDh
 *     DI
 *     LD      (5B5Ch),A
 *     OUT     (C),A
 *     EI
 *
 * The principle is the same for all bank switching; only change the bits you
 * need to.
 *
 * On the 128 and +2, Banks 1,3,4,6 and most of 7 are used for the silicon 
 * disc; the rest of 7 contains editor scratchpads. On the +2A and +3, 
 * Banks 1,3,4 and 6 are used for the disc cache and RAMdisc, while Bank 7
 * contains editor scratchpads and +3DOS workspace.
 *******************************************************************************/

/********************************************************************************
 * outBankmPl3() - Spectrum +2A and +3 memory management.
 *
 * Input:  val - control byte
 *
 * Return: -
 *******************************************************************************/

static inline void 
outBankmPl3(uns8 val)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IO)
	{
		XzxMesg(XZX_DEBUG, 
				"bankm: RAM select %d, ROM select %d, screen %d, paging %sabled",
				val & BIT_SELRAM, (val & BIT_SELROM) >> 4, (val & BIT_SELSCREEN) >> 3, 
				PAGING_DISABLED(Last0x7FFD) ? "dis" : "en");
	}
#endif

	if (Last0x7FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}

	/* First bit D3, screen select; 0 selects page 5, 1 selects page 7 */
	ScreenSelect = val & BIT_SELSCREEN ? RAM7 : RAM5;
	if ((val & BIT_SELSCREEN) != (Last0x7FFD & BIT_SELSCREEN))
	{
		ForceScreenRefresh();
	}

	/* Now bit D5, the page disable bit */
	Last0x7FFD = val;

	/* First check bits D0-D2; RAM select, could be extended */
	if (!EXTENDED_RAM(Last0x1FFD))
	{
		/* Now check bits D0-D2; RAM select, could be extended */
		PageIn(3, RAM0 + (val & BIT_SELRAM));

		/* Now bit D4, ROM select at 0x0000 in combination with bit 2 of
		   port 0x1ffd. This is the horizontal switch. */
		PageIn(0, ((Last0x1FFD & BIT_SELROM3) >> 1) | ((val & BIT_SELROM) >> 4));

#ifdef DEBUG
		if (GETCFG(debug) & D_IO)
		{
			XzxMesg(XZX_DEBUG, "bankm: horizontal ROM switch");
		}
#endif
	}
}

/********************************************************************************
 * OutBank678() - Additional memory management on the +2A and +3.
 *
 * Input:  val - control byte
 *
 * Return: -
 *******************************************************************************/

void
OutBank678(uns8 val)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IO)
	{
		XzxMesg(XZX_DEBUG, "bank678: %s", EXTENDED_RAM(Last0x1FFD) ?
				"extended RAM enabled" : "vertical ROM switch");
		}
#endif

	if (Last0x1FFD == val || PAGING_DISABLED(Last0x7FFD))
	{
		return;
	}

	if (!EXTENDED_RAM(val))
	{
		PageIn(0, ((Last0x7FFD & BIT_SELROM) >> 4) | ((val & BIT_SELROM3) >> 1));
	}
	else {
		switch ((val & BIT_SELRAM3) >> 1)
		{
			case 0:
				PageIn(0, RAM0);
				PageIn(1, RAM1);
				PageIn(2, RAM2);
				PageIn(3, RAM3);
#ifdef DEBUG
				if (GETCFG(debug) & D_IO)
				{
					XzxMesg(XZX_DEBUG, "bank678: 0,1,2,3");
				}
#endif
				break;
			case 1:
				PageIn(0, RAM4);
				PageIn(1, RAM5);
				PageIn(2, RAM6);
				PageIn(3, RAM7);
#ifdef DEBUG
				if (GETCFG(debug) & D_IO)
				{
					XzxMesg(XZX_DEBUG, "bank678: 4,5,6,7");
				}
#endif
				break;
			case 2:
				PageIn(0, RAM4);
				PageIn(1, RAM5);
				PageIn(2, RAM6);
				PageIn(3, RAM3);
#ifdef DEBUG
				if (GETCFG(debug) & D_IO)
				{
					XzxMesg(XZX_DEBUG, "bank678: 4,5,6,3");
				}
#endif
				break;
			case 3:
				PageIn(0, RAM4);
				PageIn(1, RAM7);
				PageIn(2, RAM6);
				PageIn(3, RAM3);
#ifdef DEBUG
				if (GETCFG(debug) & D_IO)
				{
					XzxMesg(XZX_DEBUG, "bank678: 4,7,6,3");
				}
#endif
				break;
		}
	}

#ifdef DEBUG
	if (GETCFG(debug) & D_IO)
	{
		if ((val & BIT_DISK) != (Last0x1FFD & BIT_DISK))
		{
			XzxMesg(XZX_DEBUG, "disk %sed", val & BIT_DISK ? "start" : "stopp");
		}
	}
#endif

	Last0x1FFD = val;
}

/********************************************************************************
 * OutByte() - Write a byte to a port.
 *
 * Input:  port - I/O port
 *         val  - byte
 *
 * Return: -
 *******************************************************************************/

void
OutByte(uns16 port, uns8 val)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IO && LOWBYTE(port) != PORT_ULA)
	{
		XzxMesg(XZX_DEBUG, "writing <%02x> to port <%04x>", val, port);
	}
#endif

	/* Border color and (possibly) speaker control */
	if (!(port & BIT_ULA))
	{
		/* Border control */
		if (LastBorderColor != (val & BIT_BORDER))
		{
			LastBorderColor = val & BIT_BORDER;
			SetBorderColor(LastBorderColor);
		}

		/* Speaker control */
// ###		SpeakerOutByte(outPorts[PORT_ULA], val);
	}

	/* Spectrum 128K, +2, +3 and 2A ports */
	if (LOWBYTE(port) == PORT_128)
	{
		switch(GETCFG(machine))
		{
			case 48:
				/*XzxMesg(XZX_WARN,
						"output to 128K/+3 port <%04x> in 48K mode", port);*/
				break;

			case 128:
				if (port == PORT_BANK128)
				{
					/* Banked memory management */
					outBankm128(val);
				}
				break;

			case 3:
				switch(port)
				{
					case PORT_BANK128:
						/* Banked memory management */
						outBankmPl3(val);
						break;
					case PORT_BANK3:
						/* Banked memory management */
						OutBank678(val);
						break;
					case PORT_CENTRONICS:
					case PORT_FDCCONTROL:
					case PORT_FDCDATA:
						XzxMesg(XZX_WARN,
								"output to unhandled +3 port <%04x>", port);
						break;
				}
				break;
		}

		/* AY-3-8912 emulation */
		if (port & BIT_SNDCHIP)
		{
			if (port & BIT_SNDCONTROL)
			{
				Last0xFFFD = val;
			}
			else
			{
				/* Store register values for snapshoting here, because we can't
				   ask the daemon for it. */
				PSG[Last0xFFFD] = val;
// ###				AyOutByte(Last0xFFFD, val);
			}
		}
	}

	/* ZX printer */
	if (!(port & BIT_PRINTER))
	{
		/* ignore */
	}

	/* Interface I microdrive I/O */
	if (GETCFG(if1_active))
	{
		if (LOWBYTE(port) == PORT_MDCONTROL)
		{
			OutPortEF(val);
		}
		else if (LOWBYTE(port) == PORT_MDDATA)
		{
			OutPortE7(val);
		}
	}

#ifdef PSEUDO_IO
	if (port == PSEUDO_STDOUT)
	{
		putc((int)val, stdout);
	}
	else if (port == PSEUDO_STDERR)
	{
		putc((int)val, stderr);
	}
#endif

	outPorts[LOWBYTE(port)] = val;
}

/********************************************************************************
 * InByte() - Read a byte from a port.
 *
 * Input:  port - I/O port
 *
 * Return: byte
 *******************************************************************************/

uns8
InByte(uns16 port)
{
#ifdef DEBUG
	if (GETCFG(debug) & D_IO && LOWBYTE(port) != PORT_ULA)
	{
		XzxMesg(XZX_DEBUG, "reading from port <%04x>", port);
	}
#endif

	/* If you read from a port that activates both the keyboard and a joystick
	   port (e.g. Kempston), the joystick takes priority - see Street Hawk and
	   Command4. */

	/* Kempston joystick */
	if (!(port & BIT_KEMPSTON))
	{
		if (GETCFG(useJoy))
		{
			InPorts[PORT_KEMPSTON] = ReadJoy();
		}
		return (InPorts[PORT_KEMPSTON]);
	}
	
	/* Keyboard and tape in */
	if (!(port & BIT_ULA))
	{
		/* Bit 5 (value 64) of IN-port FE is the ear input bit. When the line is
		   silent, its value is zero, except in the early Model 2 of the
		   Spectrum, where it was one. Some old programs rely on the fact that
		   bit 5 is always one (for instance Spinads). Bits 6 and 7 are always
		   one. */

		int res = (GETCFG(machine) == 48 && GETCFG(issue) == 2) ? 0xff : 0xdf;

		if (!(port & 0x8000))
		{
			res &= KeyPorts[7];
		}
		if (!(port & 0x4000))
		{
			res &= KeyPorts[6];
		}
		if (!(port & 0x2000))
		{
			res &= KeyPorts[5];
		}
		if (!(port & 0x1000))
		{
			res &= KeyPorts[4];
		}
		if (!(port & 0x0800))
		{
			res &= KeyPorts[3];
		}
		if (!(port & 0x0400))
		{
			res &= KeyPorts[2];
		}
		if (!(port & 0x0200))
		{
			res &= KeyPorts[1];
		}
		if (!(port & 0x0100))
		{
			res &= KeyPorts[0];
		}

		return res;
	}

	/* Spectrum 128K, +2, +3 and 2A ports */
	if (LOWBYTE(port) == PORT_128)
	{
		if (GETCFG(machine) == 48)
		{
			/*XzxMesg(XZX_WARN,
					"input from 128K/+3 port <%04x> in 48K mode", port);*/
		}

		else if (GETCFG(machine) == 3)
		{
			switch(port)
			{
				case PORT_CENTRONICS:
				case PORT_FDCCONTROL:
				case PORT_FDCDATA:
					XzxMesg(XZX_WARN,
							"input from unhandled +3 port <%04x>", port);
					return 0xff;
			}
		}

		/* AY-3-8912 emulation */
		if (port & (BIT_SNDCHIP | BIT_SNDDATA) == (BIT_SNDCHIP | BIT_SNDDATA))
		{
			return (PSG[Last0xFFFD]);
		}
		return 0xff;
	}

	/* ZX printer */
	if (!(port & BIT_PRINTER))
	{
		/* no printer present */
		return 0xff;
	}

	/* Interface I microdrive I/O */
	if (GETCFG(if1_active))
	{
		if (LOWBYTE(port) == PORT_MDCONTROL)
		{
			return (InPortEF());
		}
		else if (LOWBYTE(port) == PORT_MDDATA)
		{
			return (InPortE7());
		}
	}

	/* Multiface 128 paging */
	if (GETCFG(mf128_active))
	{
		if (LOWBYTE(port) == PORT_MF128OUT)
		{
			PageIn(0, GET_BASIC_ROM(GETCFG(machine)));
			return 0xff;
		}
		else if (LOWBYTE(port) == PORT_MF128IN)
		{
			PageIn(0, ROM3);

			/* Bit 7 represents bit 3 of the byte last written to port 7FFDh */
			return (Last0x7FFD & BIT_SELSCREEN ? 0xff : 0x7f);
		}
	}

#ifdef PSEUDO_IO
	if (port == PSEUDO_STDIN)
	{
		return ((uns8)getchar());
	}
#endif

	/* If the processor reads from a non-existing IN port, for instance FF, the
	   ULA won't stop, but nothing will put anything on the data bus. Therefore,
	   you'll read a mixture of FF's (idle bus), and screen and ATTR data bytes
	   (the latter being very scarce, by the way). This will only happen when
	   the ULA is reading the screen memory, about 60% of the 1/50th second time
	   slice in which a frame is generated. The other 40% the ULA is building
	   the border or generating a vertical retrace. This behaviour is actually
	   used in some programs, for instance, in Arkanoid. */

	return (Vline & 192 ?
			RD_BYTE(ATTR_OFFSET | ((Vline - 64) & 0xf8) << 2) : 0xff);
}

/* EOF *************************************************************************/
