/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char rcsid[] = "$Id: util.c,v 3.0 1996/10/03 10:23:45 erik Exp $";

/********************************************************************************
 * util.c - Contain various common utility functions.
 *******************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <ctype.h>

#include <OS.h>

#include "ZXConfig.h"
#include "z80.h"
#include "mem.h"
#include "main.h"
#include "util.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

/********************************************************************************
 * Forwards
 *******************************************************************************/

/********************************************************************************
 * Globals
 *******************************************************************************/

/********************************************************************************
 * Locals
 *******************************************************************************/

/* List of functions to be called when shutting down */
static struct on_quit_rec {
	void (*f)();
	struct on_quit_rec *next;
} *onQuitList = NULL;

/********************************************************************************
 * OpenXzxFile() - Open a file, checking each element in the colon-separated
 *                 resource 'libDir', unless the filename is an absolute path
 *                 or relative to the current or parent directory. Behaves
 *                 like fopen() in other respects.
 *
 * Input:  name  - name of file
 *         amode - access mode
 *
 * Return: fp - file pointer
 *         0  - error
 *******************************************************************************/

FILE *
OpenXzxFile(char *name, char *amode)
{
	FILE *fp = 0;
	char *buf;
	char *path, *element;

	if (!name || !name[0])
	{
		return NULL;
	}

	if (name[0] == '/' || !strncmp(name, "./", 2) || !strncmp(name, "../", 3))
	{
		return (fopen(name, amode));
	}

	if ((path = (char *)malloc(strlen(GETCFG(libDir)) + 1)) == 0)
	{
		XzxMesg(XZX_FATAL, "malloc failed in OpenXzxFile()");
	}

	strcpy(path, GETCFG(libDir));

	element = strtok(path, ":");
	while (element)
	{
		if ((buf = (char *)malloc(strlen(name) + strlen(element) + 2)) == 0)
		{
			XzxMesg(XZX_FATAL, "malloc failed in OpenXzxFile()");
		}

		sprintf(buf, "%s/%s", element, name);
		fp = fopen(buf, amode);
		free(buf);
		if (fp)
		{
			break;
		}
		element = strtok(NULL, ":");
	}
	free(path);

	return fp;
}

/********************************************************************************
 * OnQuit() - Insert a function address to be called during shutdown of xzx.
 *
 * Input:  fn - function address
 *
 * Return: -
 *******************************************************************************/
 
void
OnQuit(void (*fn)())
{
    struct on_quit_rec *p;

	if ((p = (struct on_quit_rec *)malloc(sizeof (struct on_quit_rec))) == 0)
	{
		XzxMesg(XZX_FATAL, "malloc failed in OnQuit()");
	}

	p->f = fn;
	p->next = onQuitList;
	onQuitList = p;
}

/********************************************************************************
 * Quit() - Shut down the emulator cleanly, calling any functions on the Quit
 *          list.
 *
 * Input:  status - xzx return code
 *
 * Return: -
 *******************************************************************************/

void
Quit(int status)
{
	struct on_quit_rec *p = onQuitList;

	while (p)
	{
		p->f();
		p = p->next;
	}

	/* We only exit the emulator thread here. */
	exit_thread(status);
}

/********************************************************************************
 * XzxMesg() - Report an info, warning, error or fatal message. Exit if a fatal
 *             message has been reported.
 *
 * Input: error_level - error level
 *        fmt         - format string for printf
 *        args        - arguents for printf
 *
 * Return: -
 *******************************************************************************/

void
XzxMesg(int errorLevel, ...)
{
	va_list args;
	char *fmt;

	va_start(args, errorLevel);
	fmt = (char *)va_arg(args, char *);

	fprintf(stderr, "%s: ", ProgName);
	switch(errorLevel)
	{
		case XZX_WARN:
		case XZX_PWARN:
			fprintf(stderr, "warning: ");
			break;
		case XZX_ERR:
		case XZX_PERR:
			fprintf(stderr, "error: ");
			break;
		case XZX_FATAL:
			fprintf(stderr, "fatal: ");
			break;
		case XZX_DEBUG:
			fprintf(stderr, "debug: ");
			break;
	}

	vfprintf(stderr, fmt, args);
	fputc('\n', stderr);

	if (errorLevel == XZX_PWARN || errorLevel == XZX_PERR)
	{
		perror("  -> reason");
	}
	else if (errorLevel == XZX_FATAL)
	{
		Quit(1);
	}
}

/********************************************************************************
 * ItimerTask() - Task function to implement interval timer in BeOS.
 *
 * Input:  data - thread data pointer
 *
 * Return: long - thread exit status
 *******************************************************************************/

/* Thread data for interval timer thread. */
typedef struct {
	bigtime_t snoozeTime;		/* interval between ticks */
	void (*pFunction)(void);	/* function called at each tick */
	bool stop;						/* flag for shutting down */
} itimerInfo;

static long
ItimerTask(void *data)
{
	itimerInfo *pInfo = (itimerInfo *)data;

	while (!pInfo->stop) {
		snooze(pInfo->snoozeTime);
		(pInfo->pFunction)();
	}

	return 0;
}

/********************************************************************************
 * IntFrequency() - Set the interrupt frequency in microseconds. 0 disables.
 *
 * Input:  ms - time interval in microseconds
 *         pFunction - a function called at each tick
 *
 * Return: -
 *******************************************************************************/

void
IntFrequency(int ms, void (*pFunction)(void))
{
	static thread_id itimerThread = B_ERROR;
	static itimerInfo info = { 0, 0, false };

	if (ms > 0 && pFunction) {
		info.snoozeTime = (bigtime_t)ms;
		info.pFunction = pFunction;
		
		/* Spawn new timer thread if necessary. */
		if (itimerThread < B_OK) {
			itimerThread = spawn_thread(ItimerTask, "Interval timer",
												 B_NORMAL_PRIORITY, &info);
			if (itimerThread >= B_OK) {
				info.stop = false;
				resume_thread(itimerThread);
			}
		}
	} else {
		/* Kill any timer thread already running. */
		if (itimerThread >= B_OK) {
			long dummy;
			info.stop = true;
			wait_for_thread(itimerThread, &dummy);
			itimerThread = B_ERROR;
		}
	}
}

/* EOF *************************************************************************/
