/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *               Jens Kilian  1998
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 * modified for BeZX JJK
 *******************************************************************************/

static char z80_c[] = "$Id: z80.c,v 3.0 1996/10/03 10:22:03 erik Exp $";

/********************************************************************************
 * z80.c - This is the heart of the Z80 emulation.
 *******************************************************************************/

#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "util.h"
#include "audio.h"
#include "if1.h"
#include "loadsave.h"
#include "snapshot.h"

/********************************************************************************
 * Defines
 *******************************************************************************/

/* This kludge is for DEC compilers, which define DEC. */
#ifdef DEC
#undef DEC
#endif

#define PARITY(reg)		parityTbl[(reg)]

/********************************************************************************
 * Forwards
 *******************************************************************************/

static void z80_Quit(void);
#ifndef HAVE_ENOUGH_SWAP
static void z80_DecodeCB(uns8);
static void z80_DecodeED(uns8);
static void z80_DecodeDD(uns8);
static void z80_DecodeFD(uns8);
#endif

/********************************************************************************
 * Globals
 *******************************************************************************/

struct z80 Cpu;					/* The Z80 cpu */
int	CarryFlag;					/* The flags of the F register */
int AddSubFlag;
int ParityOverflowFlag;
int HalfCarryFlag;
int ZeroFlag;
int SignFlag;
int Tstates = 0;				/* # of Z80 T-states */

/********************************************************************************
 * Locals
 *******************************************************************************/

static uns32 work32;			/* 32-bit temporary workspace */
static uns16 work16;			/* 16-bit temporary workspace */
static uns8	work8, work8_2;		/* 8-bit  temporary workspace */

static int idx;					/* For half-carry & overflow detection */

static uns8 high, low;			/* Used by DAA */
static uns8	add, carry,	sub_;

static int delayedEI = 0;		/* Flag and counter to carry out a delayed */
static int EIcount = 1;			/* EI as per Z80 specs */

/* Parity table - 256 bytes; 1 = even parity, 0 = odd parity */
static int parityTbl[] = {
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
};

/* Half-carry tables */
static int halfCarryTbl[]    = { 0, 0, 1, 0, 1, 0, 1, 1 };
static int subHalfCarryTbl[] = { 0, 1, 1, 1, 0, 0, 0, 1 };

/* Overflow tables */
static int overflowTbl[]     = { 0, 1, 0, 0, 0, 0, 1, 0 };
static int subOverflowTbl[]  = { 0, 0, 0, 1, 1, 0, 0, 0 };

/********************************************************************************
 * Some common macros for Z80 emulation.
 *******************************************************************************/

#define INC(reg) \
	{															\
		++(reg);												\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		H_FLAG = (((reg) & 0x0f) == 0);							\
		V_FLAG = ((reg) == 0x80);								\
		CLR(N_FLAG);											\
	}

#define DEC(reg) \
	{															\
		--(reg);												\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		H_FLAG = (((reg) & 0x0f) == 0x0f);						\
		V_FLAG = ((reg) == 0x7f);								\
		SET(N_FLAG);											\
	}

#define ADD(val) \
	{															\
		work16 = A + (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}

#define ADC(val) \
	{															\
		work16 = A + (val) + C_FLAG;							\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}

#define SUB(val) \
	{															\
		work16 = A - (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}

#define SBC(val) \
	{															\
		work16 = A - (val) - C_FLAG;							\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}

#define AND(val) \
	{															\
		A &= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		SET(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}

#define XOR(val) \
	{															\
		A ^= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		CLR(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}

#define OR(val) \
	{															\
		A |= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		CLR(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}

#define CMP(val) \
	{															\
		work16 = A - (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		S_FLAG = ((work16 & 0x80) != 0);						\
		Z_FLAG = ((work16 & 0xff) == 0);						\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}
	
#define ADD_HL(reg) \
	{															\
		work32 = HL + (reg);									\
		idx = ((HL & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		HL = work32 & 0xffff;									\
		H_FLAG = halfCarryTbl[idx];								\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}


#define ADC_HL(reg) \
	{															\
		work32 = HL + (reg) + C_FLAG;							\
		idx = ((HL & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		HL = work32 & 0xffff;									\
		S_FLAG = ((HL & 0x8000) != 0);							\
		Z_FLAG = (HL == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}

#define SBC_HL(reg) \
	{															\
		work32 = HL - (reg) - C_FLAG;							\
		idx = ((HL & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		HL = work32 & 0xffff;									\
		S_FLAG = ((HL & 0x8000) != 0);							\
		Z_FLAG = (HL == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}

#define ADD_XX(reg) \
	{															\
		work32 = XX + (reg);									\
		idx = ((XX & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		XX = work32 & 0xffff;									\
		S_FLAG = ((XX & 0x8000) != 0);							\
		Z_FLAG = (XX == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}

#define RLC(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) = ((reg) << 1) | C_FLAG;							\
		_SHIFT_TEST_((reg));									\
	}

#define RRC(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		(reg) = ((reg) >> 1) | (C_FLAG << 7);					\
		_SHIFT_TEST_((reg));									\
	}

#define RL(reg) \
	{															\
		work8_2 = (((reg) & 0x80) != 0);						\
		(reg) = ((reg) << 1) | C_FLAG;							\
		C_FLAG = work8_2;										\
		_SHIFT_TEST_((reg));									\
	}

#define RR(reg) \
	{															\
		work8_2 = (reg) & 0x01;									\
		(reg) = ((reg) >> 1) | (C_FLAG << 7);					\
		C_FLAG = work8_2;										\
		_SHIFT_TEST_((reg));									\
	}

#define SLA(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) <<= 1;											\
		_SHIFT_TEST_((reg));									\
	}

#define SRA(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		work8_2 = (reg) & 0x80;									\
		(reg) = ((reg) >> 1) | work8_2;							\
		_SHIFT_TEST_((reg));									\
	}

#define SLI(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) = ((reg) << 1) | 0x01;							\
		_SHIFT_TEST_((reg));									\
	}

#define SRL(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		(reg) >>= 1;;											\
		_SHIFT_TEST_((reg));									\
	}

#define _SHIFT_TEST_(reg) \
	{															\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		CLR(H_FLAG);											\
		P_FLAG = PARITY((reg));									\
		CLR(N_FLAG);											\
	}

#define BIT(bit, reg) \
	{															\
		Z_FLAG = (((reg) & (0x01 << (bit))) == 0);				\
		SET(H_FLAG);											\
		P_FLAG = Z_FLAG;										\
		CLR(N_FLAG);											\
	}

#define IN(port, reg) \
	{															\
		(reg) = InByte((port));									\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		CLR(H_FLAG);											\
		P_FLAG = PARITY((reg));									\
		CLR(N_FLAG);											\
	}

#define CALL \
	{															\
		PUSH(PC + 2);											\
		PC = RD_WORD(PC);										\
	}

#define RST(addr) \
	{															\
		PUSH(PC);												\
		PC = (addr);											\
	}

#define Z80_INTERRUPT \
	{															\
		if (IFF1)												\
		{														\
			/* Kludge: make sure we don't return to a halt		\
			   instruction. */									\
			if (op == 0x76)										\
			{													\
				PC++;											\
			}													\
																\
			IFF2 = IFF1;										\
			IFF1 = 0;											\
			PUSH(PC);											\
																\
			if (IM == 2)										\
			{													\
				Tstates += 19;									\
				PC = RD_WORD((I << 8) | 0xff);					\
			}													\
			else												\
			{													\
				Tstates += 13;									\
				PC = 0x0038;									\
			}													\
		}														\
	}

#define Z80_M1(op) \
	{															\
		(op) = RD_BYTE(PC++);									\
																\
		/* The R register is a counter that is updated every	\
		   instruction, where DD, FD, ED and CB are to be		\
		   regarded as separate instructions. So shifted		\
		   instruction will increase R by two. There's an		\
		   interesting exception: doubly-shifted opcodes, the	\
		   DDCB and FDCB ones, increase R by two too. LDI		\
		   increases R by two, LDIR increases it by 2 times BC,	\
		   as does LDDR etcetera. The sequence LD R,A/LD A,R	\
		   increases A by two, except for the highest bit: this	\
		   bit of the R register is never changed. */			\
																\
		R = (R & 0x80) | (++R & 0x7f);							\
	}

#define Z80_DELAYED_EI \
	{															\
		/* Sort out the EI instruction - enable the interrupts	\
		   AFTER the next instruction. */						\
		if (delayedEI && !EIcount--)							\
		{														\
			IFF1 = 1;											\
			IFF2 = 1;											\
			delayedEI = 0;										\
			EIcount = 1;										\
		}														\
	}

/********************************************************************************
 * Z80Init() - Initialize the Z80 cpu emulation.
 *
 * Input:  -
 *
 * Return: 0 - o.k.
 *******************************************************************************/

int
Z80_Init(void)
{
	OnQuit(z80_Quit);
	
	S_FLAG = 0;
	Z_FLAG = 0;
	H_FLAG = 0;
	P_FLAG = 0;
	N_FLAG = 0;
	C_FLAG = 0;

	Z80_Reset();
	
	return 0;
}

/********************************************************************************
 * z80_Quit() - Shutdown the Z80 emulation.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

static void
z80_Quit(void)
{
	/* nothing todo here */
}

/********************************************************************************
 * StoreFlags() - Since the flags are stored in ints instead of in the F
 *                register itself, we must store/retrieve them whenever the F
 *                register is read/written. Fortunately, this doesn't happen
 *                much.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
Z80_StoreFlags(void)
{
	if (S_FLAG) F |= F_NEG;      else F &= ~F_NEG;
	if (Z_FLAG) F |= F_ZERO;     else F &= ~F_ZERO;
	if (H_FLAG) F |= F_HCARRY;   else F &= ~F_HCARRY;
	if (P_FLAG) F |= F_OVERFLOW; else F &= ~F_OVERFLOW;
	if (N_FLAG) F |= F_NADD;     else F &= ~F_NADD;
	if (C_FLAG) F |= F_CARRY;    else F &= ~F_CARRY;
}

/********************************************************************************
 * RetrieveFlags() - Since the flags are stored in ints instead of in the F
 *                   register itself, we must store/retrieve them whenever the F
 *                   register is read/written. Fortunately, this doesn't happen
 *                   much.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
Z80_RetrieveFlags(void)
{
	S_FLAG = F & F_NEG      ? 1 : 0;
	Z_FLAG = F & F_ZERO     ? 1 : 0;
	H_FLAG = F & F_HCARRY   ? 1 : 0;
	P_FLAG = F & F_OVERFLOW ? 1 : 0;
	N_FLAG = F & F_NADD     ? 1 : 0;
	C_FLAG = F & F_CARRY    ? 1 : 0;
}

/********************************************************************************
 * ResetCPU() - Reset the Z80 cpu.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
Z80_Reset(void)
{
	int i;
	
	PC   = 0;
	I    = 0;
	R    = 0;
	IM   = 0;
	IFF1 = 0;
	IFF2 = 0;

	*LastSnapshotName = 0;

	InitPages();

	Last0x7FFD = 0;
	Last0x1FFD = 0;
	Last0xFFFD = 0;

	if (GETCFG(machine) != 48)
	{
		for (i = 0; i < sizeof(PSG); i++)
		{
			OutByte(PORT_SNDCONTROL, i);
			OutByte(PORT_SNDDATA, 0);
		}
	}
}

/********************************************************************************
 * NmiPU() - Generate an NMI.
 *
 * Input:  -
 *
 * Return: -
 *******************************************************************************/

void
Z80_NMI(void)
{
	/* When an NMI occurs, the total time taken to get to address 66h is
	   15 cycles. This is the same as for an ordinary interrupt but with
	   two extra cycles (reason unknown). */

	Tstates += 15;

	IFF2 = IFF1;
	IFF1 = 0;

	if (RD_BYTE(PC) == 0x76)
	{
		PC++;
	}

	PUSH(PC);
	PC = 0x0066;
}

/********************************************************************************
 * z80_DecodeXX() - This is a kludge for machines that run out of swapspace when
 *                  compiling this file. You should blame your system administra-
 *                  tor and not me for this!
 *
 * Input:  op - Z80 op code
 *
 * Return: -
 *******************************************************************************/

#ifndef HAVE_ENOUGH_SWAP

static void
z80_DecodeCB(uns8 op)
{
	switch(op)
	{
#		include "z80_cbops.c"
	}
}

static void
z80_DecodeED(uns8 op)
{
	switch(op)
	{
#		include "z80_edops.c"
	}
}

static void
z80_DecodeDD(uns8 op)
{
	switch(op)
	{
#		define XX  IX
#		define LXX LIX
#		define HXX HIX
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
	}
}

static void
z80_DecodeFD(uns8 op)
{
	switch(op)
	{
#		define NO_RCSID
#		define XX  IY
#		define LXX LIY
#		define HXX HIY
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
#		undef NO_RCSID
	}
}

#endif /* HAVE_ENOUGH_SWAP */

/* EOF *************************************************************************/
