//	Copyright (c) Jens Kilian 1998
//
//	Permission to use, copy, modify, distribute, and sell this software and its
//	documentation for any purpose is hereby granted without fee, provided that
//	the above copyright notice appear in all copies and that both that
//	copyright notice and this permission notice appear in supporting
//	documentation, and that the name of the copyright holder not be used in
//	advertising or publicity pertaining to distribution of the software without
//	specific, written prior permission.  The copyright holder makes no
//	representations about the suitability of this software for any purpose.  It
//	is provided "as is" without express or implied warranty.
//
//	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
//	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
//	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
//	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
//	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
//	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//	PERFORMANCE OF THIS SOFTWARE.

//	The application class for BeZX.

#include <cstring>

#include <Alert.h>
#include <Application.h>
#include <Entry.h>
#include <Path.h>

#include "ZXApplication.h"
#include "ZXConfig.h"
#include "ZXKeyboardView.h"
#include "ZXKeyboardWindow.h"
#include "ZXWindow.h"

// Apllication signature.
const char kSignature[] = "application/x-jjk-bezx";

// Title of ZX Spectrum keyboard window.
const char kKeyboardWindowTitle[] = "ZX Spectrum Keyboard";

ZXApplication::ZXApplication(void)
:	inherited(kSignature),
	mArgc(0),
	mppArgv(0),
	mpKeyboardWindow(0)
{
	// Initialize dummy command-line arguments.
	const char *kDummyArgs[] = { "BeZX" };
	CopyArgs(1, kDummyArgs);
}

ZXApplication::~ZXApplication(void)
{
	delete ZXConfig::smpConf;
	ZXConfig::smpConf = 0;
}

void
ZXApplication::AboutRequested(void)
{
	(new BAlert("About BeZX",
					"BeZX - Sinclair ZX Spectrum emulator for the BeBox\n"
					"Based on XZX by Des Herriott - ported by Jens Kilian\n"
					"Copyright © 1993-1994  Des Herriott\n"
					"Copyright © 1996,1997  Jens Kilian",
					"OK"))->Go();
}

void
ZXApplication::ArgvReceived(int32 argc, char **ppArgv)
{
	// Remember command-line arguments.
	CopyArgs(argc, ppArgv);
}

void
ZXApplication::MessageReceived(BMessage *pMessage)
{
	switch (pMessage->what) {
	
	case kToggleKeyboardWindow:
		// Show resp. hide the keyboard window.
		if (mpKeyboardWindow) {
			if (mpKeyboardWindow->IsHidden()) {
				mpKeyboardWindow->Show();
				SETCFG(kbdShown, true);
			} else {
				mpKeyboardWindow->Hide();
				SETCFG(kbdShown, false);
			}
		}
		break;
		
	default:
		inherited::MessageReceived(pMessage);
		break;
	}
}

bool
ZXApplication::QuitRequested(void)
{
	// Get rid of the keyboard window.
	if (mpKeyboardWindow && mpKeyboardWindow->Lock()) {
		mpKeyboardWindow->Quit();
	}
	mpKeyboardWindow = 0;
	
	return inherited::QuitRequested();
}

void
ZXApplication::ReadyToRun(void)
{
	// Get application directory.
	BPath appPath;
	const char *pAppDir = 0;
	app_info appInfo;
	if (GetAppInfo(&appInfo) >= B_OK) {
		BEntry appEntry(&appInfo.ref);
		if (appEntry.GetParent(&appEntry) >= B_OK
			 && appEntry.GetPath(&appPath) >= B_OK) {
			pAppDir = appPath.Path();
		}
	}
	
	// Parse command-line options.
	ZXConfig::smpConf = new ZXConfig(mArgc, mppArgv, pAppDir);

	// Create the ZX Spectrum keyboard display.
	ZXKeyboardView *pKeyboardView = new ZXKeyboardView;	
	mpKeyboardWindow = new ZXKeyboardWindow(GETCFG(kbdLocation),
														 GETCFG(kbdSize),
														 pKeyboardView,
														 kKeyboardWindowTitle);
	if (GETCFG(kbdShown)) {
		mpKeyboardWindow->Show();
	}
	
	// Create and show the emulator window.
	// This will start the working threads.
	(new ZXWindow(GETCFG(location), mArgc, mppArgv))->Show();
}

void
ZXApplication::CopyArgs(int32 argc, const char **ppArgv)
{
	for (int32 i = 0; i < mArgc; ++i) {
		delete [] mppArgv[i];
	}
	delete [] mppArgv;
	
	mArgc = argc;
	mppArgv = new char *[mArgc + 1];
	for (int32 i = 0; i < mArgc; ++i) {
		mppArgv[i] = strcpy(new char[strlen(ppArgv[i])+1], ppArgv[i]);
	}
	mppArgv[mArgc] = 0;
}

//	The usual highly exciting Be main program.
int
main(void)
{
	ZXApplication theApp;
	theApp.Run();
	return 0;
}
