//	Copyright (c) Jens Kilian 1998
//
//	Permission to use, copy, modify, distribute, and sell this software and its
//	documentation for any purpose is hereby granted without fee, provided that
//	the above copyright notice appear in all copies and that both that
//	copyright notice and this permission notice appear in supporting
//	documentation, and that the name of the copyright holder not be used in
//	advertising or publicity pertaining to distribution of the software without
//	specific, written prior permission.  The copyright holder makes no
//	representations about the suitability of this software for any purpose.  It
//	is provided "as is" without express or implied warranty.
//
//	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
//	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
//	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
//	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
//	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
//	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//	PERFORMANCE OF THIS SOFTWARE.

// File selection dialog management for BeZX.

#include <cstring>
#include <cstdlib>

#include <Directory.h>
#include <Entry.h>
#include <FilePanel.h>
#include <Looper.h>
#include <Message.h>
#include <Messenger.h>
#include <OS.h>
#include <Path.h>
#include <Window.h>

#include "ZXFileDialog.h"

// Static class members.
entry_ref ZXFileDialog::smCurrentDirectory;

ZXFileDialog::ZXFileDialog(file_panel_mode mode, const char *pLabel)
:	inherited("File selection"),
	mFilePanel(mode, &BMessenger(this), CurrentDirectory(), B_FILE_NODE, false),
	mSelectedEntry(),
	mFinished(create_sem(0, "File selection"))
{
	mFilePanel.Window()->SetTitle(pLabel);
}

ZXFileDialog::~ZXFileDialog(void)
{
	mFilePanel.GetPanelDirectory(CurrentDirectory());
	
	if (mFinished >= B_OK) {
		delete_sem(mFinished);
	}
}

void
ZXFileDialog::MessageReceived(BMessage *pMessage)
{
	switch (pMessage->what) {
	
	case B_REFS_RECEIVED:
		{	entry_ref ref;
			if (pMessage->FindRef("refs", &ref) >= B_OK) {
				mSelectedEntry.SetTo(&ref);
			}
		}
		break;
	
	case B_SAVE_REQUESTED:
		{	entry_ref ref;
			char *pName;
			if (pMessage->FindRef("directory", &ref) >= B_OK
				 && pMessage->FindString("name", &pName) >= B_OK) {
				BDirectory dir(&ref);
				mSelectedEntry.SetTo(&dir, pName);
			}
		}
		break;

	case B_CANCEL:
		if (mFinished >= B_OK) {
			release_sem(mFinished);
		}
		break;
	
	default:
		inherited::MessageReceived(pMessage);
		break;
	}
}

bool
ZXFileDialog::QuitRequested(void)
{
	return false;
}

char *
ZXFileDialog::Go(void)
{
	char *pResult = 0;

	if (mFinished >= B_OK && IsLocked()) {
		Run();

		mFilePanel.Show();
		while (acquire_sem(mFinished) == B_INTERRUPTED)
			;
		
		BPath path;
		if (mSelectedEntry.InitCheck() >= B_OK
			 && mSelectedEntry.GetPath(&path) >= B_OK) {
			pResult = strdup(path.Path());
		}
	}

	if (Lock()) {
		Quit();
	}
	
	return pResult;
}

// Return an entry_ref for the current directory.
entry_ref *
ZXFileDialog::CurrentDirectory(void) const
{
	static bool initialized = false;
	if (!initialized) {
		BPath currentDirectory = ".";
		get_ref_for_path(currentDirectory.Path(), &smCurrentDirectory);
		initialized = true;
	}
	
	return &smCurrentDirectory;
}
