//	Copyright (c) Jens Kilian 1998
//
//	Permission to use, copy, modify, distribute, and sell this software and its
//	documentation for any purpose is hereby granted without fee, provided that
//	the above copyright notice appear in all copies and that both that
//	copyright notice and this permission notice appear in supporting
//	documentation, and that the name of the copyright holder not be used in
//	advertising or publicity pertaining to distribution of the software without
//	specific, written prior permission.  The copyright holder makes no
//	representations about the suitability of this software for any purpose.  It
//	is provided "as is" without express or implied warranty.
//
//	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
//	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
//	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
//	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
//	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
//	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//	PERFORMANCE OF THIS SOFTWARE.

//	The view class for BeZX.

#include <Application.h>
#include <Bitmap.h>
#include <Message.h>
#include <OS.h>
#include <Rect.h>
#include <Region.h>
#include <View.h>

#include "ZXApplication.h"
#include "ZXConfig.h"
#include "ZXView.h"
#include "io.h"
#include "screen.h"

//	Table mapping ASCII keys to a keyboard row and bit to set/reset.
//	E.g. keyTable['a'] means set/reset bit 0 (0xfe) of keyboard row 1.
static struct {
	int port;	// which keyboard row is the key on
	int bit;		// which bit to mask
} keyTable[128] = {
	{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	// SP       !        "        #        $        %        &        '   
	{7,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xf7},{8,0x00},
	// (        )        *        +        ,        -        .        /   
	{4,0xfd},{4,0xfe},{4,0xfb},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	// 0        1        2        3        4        5        6        7   
	{4,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xef},{4,0xf7},
	// 8        9	     :        ;        <        =        >        ?   
	{4,0xfb},{4,0xfd},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00},
	// @        A        B        C        D        E        F        G   
	{3,0xfd},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
	// H        I        J        K        L        M        N        O   
	{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
	// P        Q        R        S        T        U        V        W   
	{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
	// X        Y        Z        [        \        ]        ^        _   
	{0,0xfb},{5,0xef},{0,0xfd},{8,0x00},{8,0x00},{8,0x00},{4,0xef},{8,0x00},
	// `        a        b        c        d        e        f        g   
	{8,0x00},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
	// h        i        j        k        l        m        n        o   
	{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
	// p        q        r        s        t        u        v        w   
	{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
	// x        y        z        (        |        )        ~       DEL  
	{0,0xfb},{5,0xef},{0,0xfd},{8,0x00},{8,0x00},{8,0x00},{8,0x00},{8,0x00}
};

ZXView::ZXView(BRect frame, const char *pName)
:	inherited(frame, pName, B_FOLLOW_NONE, B_WILL_DRAW|B_PULSE_NEEDED),
	mZXScreenFrame(),
	mBorderRegion(),
	mpBuffer(0),
	mUnshiftCursorKeys(false),
	mEventPort(create_port(100, "Emulator event port"))
{
	// Set up a region including the Spectrum border, then create the
	// buffer bitmap for the screen.
	mBorderRegion.Include(frame);
	frame.InsetBy(BORDER_WIDTH, BORDER_WIDTH);
	mBorderRegion.Exclude(frame);

	mZXScreenFrame = frame;
	frame.OffsetBy(-BORDER_WIDTH, -BORDER_WIDTH);
	mpBuffer = new BBitmap(frame, B_COLOR_8_BIT);

	// No keys initially being pressed.
	for (int i = 0; i < 8; i++) {
		KeyPorts[i] = 0xff;
	}
}

ZXView::~ZXView(void)
{
	delete mpBuffer;
	mpBuffer = 0;
	delete_port(mEventPort);
}

void
ZXView::AttachedToWindow(void)
{
	inherited::AttachedToWindow();

	// We will paint every pixel in the view.
	SetViewColor(B_TRANSPARENT_32_BIT);
	
	// The high color serves as the Spectrum border color.
	SetHighColor(0, 0, 0);
	
	// We want input.
	MakeFocus();
}

//	Redraw (part of) the view.
void
ZXView::Draw(BRect updateRect)
{
	// Draw the border.
	FillRegion(&mBorderRegion);
	
	// Draw the Spectrum display file.
	BRect dest = updateRect & mZXScreenFrame;
	if (dest.IsValid()) {
		BRect source = dest;
		source.OffsetBy(-BORDER_WIDTH, -BORDER_WIDTH);
		DrawBitmapAsync(mpBuffer, source, dest);
	}
	Flush();
}

//	Redraw the view after the frame buffer has been updated.
//	Called from a concurrent thread; the window must be locked.
void
ZXView::Refresh(BRect area)
{
	BWindow *pWindow = Window();
	if (pWindow->Lock()) {
		BRect dest = area;
		dest.OffsetBy(BORDER_WIDTH, BORDER_WIDTH);
		DrawBitmapAsync(mpBuffer, area, dest);
		Flush();
		pWindow->Unlock();
	}
}

//	Change the simulated border color.
//	Called from a concurrent thread; the window must be locked.
void
ZXView::SetBorderColor(uint8 r, uint8 g, uint8 b)
{
	BWindow *pWindow = Window();
	if (pWindow->Lock()) {
		SetHighColor(r, g, b);
		FillRegion(&mBorderRegion);
		pWindow->Unlock();
	}
}

// Process a key-down event.
void
ZXView::KeyDown(const char *pBytes, int32 numBytes)
{
	// Retrieve raw keycode and modifier mask.
	BMessage *pMsg = Window()->CurrentMessage();
	int32 keyCode, modifiers;
	if (pMsg) {
		if (pMsg->FindInt32("key", &keyCode) < B_OK) {
			keyCode = 0;
		}
		if (pMsg->FindInt32("modifiers", &modifiers) < B_OK) {
			modifiers = 0;
		}
	}
		
	// First check for the 'action keys'.
	if (keyCode == GETCFG(actionKeys[KM_JOY_UP])) {
		InPorts[PORT_KEMPSTON] |= BIT_UP;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_DOWN])) {
		InPorts[PORT_KEMPSTON] |= BIT_DOWN;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_LEFT])) {
		InPorts[PORT_KEMPSTON] |= BIT_LEFT;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_RIGHT])) {
		InPorts[PORT_KEMPSTON] |= BIT_RIGHT;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_FIRE])) {
		InPorts[PORT_KEMPSTON] |= BIT_FIRE;
#ifdef SLOWDOWN
	} else if (keyCode == GETCFG(actionKeys[KM_KEY_SLOWER])) {
		++GETCFG(slowdown);
	} else if (keyCode == GETCFG(actionKeys[KM_KEY_FASTER])) {
		if (GETCFG(slowdown)) {
			--GETCFG(slowdown);
		}
#endif
	}

	switch (*pBytes) {

	// Emulator control keys
	case B_FUNCTION_KEY:
		switch (keyCode) {

		case B_F1_KEY:	// Show keyboard window
			be_app->PostMessage(ZXApplication::kToggleKeyboardWindow);
			break;
			
		case B_F2_KEY:	// Generate an NMI
			write_port(mEventPort, kEmulatorNMI, 0, 0);
			break;
			
		case B_F3_KEY:	// Save snapshot
			write_port(mEventPort, kSaveSnapshot, 0, 0);
			break;
			
		case B_F4_KEY:	// Load snapshot
			write_port(mEventPort, kLoadSnapshot, 0, 0);
			break;

		case B_F5_KEY:	// Turn sound on/off
			write_port(mEventPort, kAudioToggle, 0, 0);
			break;

		case B_F6_KEY:	// Turn debugging on/off
			write_port(mEventPort, kDebugToggle, 0, 0);
			break;

		case B_F7_KEY:	// Insert new floppy disk A
			write_port(mEventPort, kInsertDiskA, 0, 0);
			break;

		case B_F8_KEY:	// Insert new floppy disk B
			write_port(mEventPort, kInsertDiskB, 0, 0);
			break;

		case B_F9_KEY:	// Switch memory model
			write_port(mEventPort, kSelectModel, 0, 0);
			break;
			
		case B_F10_KEY: // Save screen dump
			write_port(mEventPort, kScreenDump, 0, 0);
			break;
		
		case B_F11_KEY: // Reset emulator
			write_port(mEventPort, kEmulatorReset, 0, 0);
			break;
			
		case B_F12_KEY: // Leave the emulator
			write_port(mEventPort, kEmulatorQuit, 0, 0);
			break;
		}
		break;
			
	case B_TAB:
		mUnshiftCursorKeys = !mUnshiftCursorKeys;
		break;

	// Ordinary Speccy keys
	case B_RETURN:
		KeyPorts[6] &= 0xfe;				// ENTER
		break;

	case B_BACKSPACE:
	case B_DELETE:
		KeyPorts[0] &= 0xfe;				// DELETE
		KeyPorts[4] &= 0xfe;
		break;
		
	case B_ESCAPE:
		KeyPorts[0] &= 0xfe;				// EDIT
		KeyPorts[3] &= 0xfe;
		break;
		
	case B_UP_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] &= 0xfe;
		}
		KeyPorts[4] &= 0xf7;
		break;
		
	case B_DOWN_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] &= 0xfe;
		}
		KeyPorts[4] &= 0xef;
		break;
		
	case B_LEFT_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] &= 0xfe;
		}
		KeyPorts[3] &= 0xef;
		break;
		
	case B_RIGHT_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] &= 0xfe;
		}
		KeyPorts[4] &= 0xfb;
		break;
	
	case '`':
		KeyPorts[7] &= 0xfd;
		KeyPorts[0] &= 0xfb;
		break;
	
	// case '~':
	//   Requires E cursor.
	
	case '-':
		KeyPorts[7] &= 0xfd;
		KeyPorts[6] &= 0xf7;
		break;
		
	case '_':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[4] &= 0xfe;
		break;
		
	case '=':
		KeyPorts[7] &= 0xfd;
		KeyPorts[6] &= 0xfd;
		break;
		
	case '+':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[6] &= 0xfb;
		break;
		
	// case '\\':
	//   Requires E cursor.
	
	// case '|':
	//   Requires E cursor.
	
	// case '[':
	//   Requires E cursor.
	
	// case '{':
	//   Requires E cursor.
	
	// case ']':
	//   Requires E cursor.
	
	// case '}':
	//   Requires E cursor.
	
	case ';':
		KeyPorts[7] &= 0xfd;
		KeyPorts[5] &= 0xfd;
		break;
		
	case ':':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[0] &= 0xfd;
		break;
		
	case '\'':
		KeyPorts[7] &= 0xfd;
		KeyPorts[4] &= 0xf7;
		break;
		
	case '"':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[5] &= 0xfe;
		break;
		
	case ',':
		KeyPorts[7] &= 0xfd;
		KeyPorts[7] &= 0xf7;
		break;
		
	case '<':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[2] &= 0xf7;
		break;
		
	case '.':
		KeyPorts[7] &= 0xfd;
		KeyPorts[7] &= 0xfb;
		break;
		
	case '>':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[2] &= 0xef;
		break;
		
	case '/':
		KeyPorts[7] &= 0xfd;
		KeyPorts[0] &= 0xef;
		break;
		
	case '?':
		KeyPorts[0] |= 0x01;
		KeyPorts[7] &= 0xfd;
		KeyPorts[0] &= 0xf7;
		break;
		
	case '*':
		KeyPorts[7] &= 0xfd;
		KeyPorts[7] &= 0xef;
		break;
		
	case '\\':
	case '|':
		KeyPorts[0] = KeyPorts[1] = KeyPorts[2] = KeyPorts[3] =
		KeyPorts[4] = KeyPorts[5] = KeyPorts[6] = KeyPorts[7] = 0xff;
		mUnshiftCursorKeys = 0;
		break;
		
	default:
		if (numBytes == 1) {
			KeyPorts[keyTable[*pBytes].port] &= keyTable[*pBytes].bit;
		}
		break;
	}
	
	// Handle modifier keys.
	ProcessModifiers((uint32)modifiers);
}

// Process a key-up event.
void
ZXView::KeyUp(const char *pBytes, int32 numBytes)
{
	// Retrieve raw keycode and modifier mask.
	BMessage *pMsg = Window()->CurrentMessage();
	int32 keyCode, modifiers;
	if (pMsg) {
		if (pMsg->FindInt32("key", &keyCode) < B_OK) {
			keyCode = 0;
		}
		if (pMsg->FindInt32("modifiers", &modifiers) < B_OK) {
			modifiers = 0;
		}
	}
		
	// First check for the 'action keys'.
	if (keyCode == GETCFG(actionKeys[KM_JOY_UP])) {
		InPorts[PORT_KEMPSTON] &= ~BIT_UP;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_DOWN])) {
		InPorts[PORT_KEMPSTON] &= ~BIT_DOWN;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_LEFT])) {
		InPorts[PORT_KEMPSTON] &= ~BIT_LEFT;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_RIGHT])) {
		InPorts[PORT_KEMPSTON] &= ~BIT_RIGHT;
	} else if (keyCode == GETCFG(actionKeys[KM_JOY_FIRE])) {
		InPorts[PORT_KEMPSTON] &= ~BIT_FIRE;
	}

	switch (*pBytes) {

	// Ordinary Speccy keys
	case B_RETURN:
		KeyPorts[6] |= 0x01;			// ENTER
		break;
		
	case B_BACKSPACE:
	case B_DELETE:
		KeyPorts[0] |= 0x01;			// DELETE
		KeyPorts[4] |= 0x01;
		break;
		
	case B_ESCAPE:
		KeyPorts[0] |= 0x01;			// EDIT
		KeyPorts[3] |= 0x01;
		break;
		
	case B_UP_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] |= 0x01;
		}
		KeyPorts[4] |= 0x08;
		break;
		
	case B_DOWN_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] |= 0x01;
		}
		KeyPorts[4] |= 0x10;
		break;
		
	case B_LEFT_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] |= 0x01;
		}
		KeyPorts[3] |= 0x10;
		break;
		
	case B_RIGHT_ARROW:
		if (!mUnshiftCursorKeys) {
			KeyPorts[0] |= 0x01;
		}
		KeyPorts[4] |= 0x04;
		break;
		
	case '`':
		KeyPorts[7] |= 0x02;
		KeyPorts[0] |= 0x04;
		break;
		
	// case '~':
	//   Requires E cursor.
	
	case '-':
		KeyPorts[7] |= 0x02;
		KeyPorts[6] |= 0x08;
		break;
		
	case '_':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[4] |= 0x01;
		break;
		
	case '=':
		KeyPorts[7] |= 0x02;
		KeyPorts[6] |= 0x02;
		break;
		
	case '+':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[6] |= 0x04;
		break;
		
	// case '\\':
	//   Requires E cursor.
	
	// case '|':
	//   Requires E cursor.
	
	// case '[':
	//   Requires E cursor.
	
	// case '{':
	//   Requires E cursor.
	
	// case ']':
	//   Requires E cursor.
	
	// case '}':
	//   Requires E cursor.
	
	case ';':
		KeyPorts[7] |= 0x02;
		KeyPorts[5] |= 0x02;
		break;
		
	case ':':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[0] |= 0x02;
		break;
		
	case '\'':
		KeyPorts[7] |= 0x02;
		KeyPorts[4] |= 0x08;
		break;
		
	case '"':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[5] |= 0x01;
		break;
		
	case ',':
		KeyPorts[7] |= 0x02;
		KeyPorts[7] |= 0x08;
		break;
		
	case '<':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[2] |= 0x08;
		break;
		
	case '.':
		KeyPorts[7] |= 0x02;
		KeyPorts[7] |= 0x04;
		break;
		
	case '>':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[2] |= 0x10;
		break;
		
	case '/':
		KeyPorts[7] |= 0x02;
		KeyPorts[0] |= 0x10;
		break;
		
	case '?':
		KeyPorts[0] &= 0xfe;
		KeyPorts[7] |= 0x02;
		KeyPorts[0] |= 0x08;
		break;
		
	case '*':
		KeyPorts[7] |= 0x02;
		KeyPorts[7] |= 0x10;
		break;
		
	case '\\':
	case '|':
		KeyPorts[0] = KeyPorts[1] = KeyPorts[2] = KeyPorts[3] =
		KeyPorts[4] = KeyPorts[5] = KeyPorts[6] = KeyPorts[7] = 0xff;
		break;
		
	default:
		if (numBytes == 1) {
			KeyPorts[keyTable[*pBytes].port] |= ~keyTable[*pBytes].bit;
		}
		break;
	}
	
	// Handle modifier keys.
	ProcessModifiers((uint32)modifiers);
}

// Periodically check the modifier keys (which aren't reported separately).
void
ZXView::Pulse(void)
{
	if (Window()->IsActive()) {
		ProcessModifiers(modifiers());
	}
}

// Process modifier keys.
void
ZXView::ProcessModifiers(uint32 modifiers)
{
	static uint32 oldModifiers = 0;
	const uint32 pressed = ~oldModifiers & modifiers;
	const uint32 released = oldModifiers & ~modifiers;
	oldModifiers = modifiers;

	if (pressed & B_LEFT_SHIFT_KEY) {
		KeyPorts[0] &= 0xfe;
	} else if (released & B_LEFT_SHIFT_KEY) {
		KeyPorts[0] |= 0x01;
	}
	
	if (pressed & B_RIGHT_SHIFT_KEY) {
		KeyPorts[7] &= 0xfd;
	} else if (released & B_RIGHT_SHIFT_KEY) {
		KeyPorts[7] |= 0x02;
	}
}
