//	Copyright (c) Jens Kilian 1998
//
//	Permission to use, copy, modify, distribute, and sell this software and its
//	documentation for any purpose is hereby granted without fee, provided that
//	the above copyright notice appear in all copies and that both that
//	copyright notice and this permission notice appear in supporting
//	documentation, and that the name of the copyright holder not be used in
//	advertising or publicity pertaining to distribution of the software without
//	specific, written prior permission.  The copyright holder makes no
//	representations about the suitability of this software for any purpose.  It
//	is provided "as is" without express or implied warranty.
//
//	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
//	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
//	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
//	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
//	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
//	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
//	PERFORMANCE OF THIS SOFTWARE.

// The window class for BeZX.

#include <Application.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <Window.h>

#include "ZXConfig.h"
#include "ZXView.h"
#include "ZXWindow.h"
#include "main.h"
#include "screen.h"

static BRect
windowFrame(BPoint where)
{
	const float width = 2*BORDER_WIDTH + HSize;
	const float height = 2*BORDER_WIDTH + VSize;
	return BRect(where, where + BPoint(width-1, height-1));
}

ZXWindow::ZXWindow(BPoint location, int32 argc, char **ppArgv)
:	inherited(windowFrame(location),
				 (argc > 0 ? ppArgv[0] : "BeZX"),
				 B_TITLED_WINDOW, B_NOT_RESIZABLE|B_NOT_ZOOMABLE),
	mArgc(argc),
	mppArgv(ppArgv),
	mEmulatorSem(B_ERROR),
	mEmulatorThread(B_ERROR)
{
	// Set the pulse rate (needed by ZXView).
	SetPulseRate((bigtime_t)100000);
	
	// Create the view and attach it.
	ZXView *pView = new ZXView(Bounds(), "BeZX view");
	AddChild(pView);
	InitScreen(*pView);
	
	// Start up the emulator thread.
	mEmulatorSem = create_sem(1, "Emulator interlock");
	if (mEmulatorSem >= B_OK) {
		mEmulatorThread = spawn_thread(Emulate, "Emulation",
												 B_NORMAL_PRIORITY, this);
		if (mEmulatorThread >= B_OK) {
			if (resume_thread(mEmulatorThread) < B_OK) {
				kill_thread(mEmulatorThread);
				mEmulatorThread = B_ERROR;
			}
		}
	
		if (mEmulatorThread < B_OK) {
			delete_sem(mEmulatorSem);
			mEmulatorSem = B_ERROR;
		}
	}
}

ZXWindow::~ZXWindow(void)
{
	// empty
}

void
ZXWindow::FrameMoved(BPoint location)
{
	// Remember current position.
	SETCFG(location, location);
}

bool
ZXWindow::QuitRequested(void)
{
	// Shut down the emulator thread.
	if (mEmulatorSem >= B_OK) {
		acquire_sem(mEmulatorSem);
		if (mEmulatorThread >= B_OK) {
			kill_thread(mEmulatorThread);
			mEmulatorThread = B_ERROR;
		}
		release_sem(mEmulatorSem);

		delete_sem(mEmulatorSem);
		mEmulatorSem = B_ERROR;
	}
	
	// OK, tell the application to shut itself down.
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

int32
ZXWindow::Emulate(void *pData)
{
	ZXWindow *pThis = (ZXWindow *)pData;
	return pThis->Emulate();
}

int32
ZXWindow::Emulate(void)
{
	// Call emulator main program.
	zx_main(mArgc, mppArgv);
	
	// Emulator stopped running; clean up and quit.
	acquire_sem(mEmulatorSem);
	mEmulatorThread = B_ERROR;
	PostMessage(B_QUIT_REQUESTED);
	release_sem(mEmulatorSem);
	
	return 0;
}
